#ifndef _GOLDEN_CONN_H_
#define _GOLDEN_CONN_H_

#include <string>

#define KW_DATA_END 1

using namespace std;

/** Abstract class for Connection objects to connect to different database 
  * vendors. For MySQL, Oracle or any other database, only the vendor
  * object to be defined by inheriting this object
  */
class GoldenConn
{
public:
   /** Constructor to be called from derived object
     * \param db_name Name of database
     * \param host Name or ip address of machine where database is stored
     * \param user Name of user who has permission to database
     * \param passwd Password for the user
     * \param port_num Port number to the database
     */
   GoldenConn
   (
      const char  *db_name,
      const char  *host,
      const char  *user,
      const char  *passwd,
      unsigned int port_num
   )
   {
      dbName = db_name;
      hostName = host;
      userName = user;
      passWord = passwd;
      portNumber = port_num;
      numberOfRows = 0L;
      dataFlag = 0;
   }

   /** Virtual function for connecting to database. The function must be
     * implemented in the derived class
     * \return Returns 0 on succes else greater than 0
     */
   virtual int connect() = 0;
   virtual ~GoldenConn() {}

   /** Function to be implemented in the derived class for executing the
     * query on the database of a particular vendor
     * \param qry Query to be executed
     * \param store_result Flag to indicate whether to store the result or not
     * \return Returns 0 on success else greater than 0
     */
   virtual int executeQuery(const char *qry, bool store_result = true) = 0;
   
   /** Function to be implemented in the derived class for getting the next row
     * from database
     * \return Returns the row data else NULL
     */
   virtual char **nextRow() = 0;
   /** Function to be implemented in the derived class for getting the next card
     * from database. Getting card fetches the next data line for the keyword
     * \return Returns the row data else NULL
     */
   virtual char **nextCard() = 0;
   /** Function to be implemented in the derived class for getting the error
     * occurred in the database
     * \return Returns the error message if one exists
     */
   virtual const char *error() = 0;
   /** Function to be implemented in the derived class for getting the number
     * of rows after the execution of the query in the result set
     * \return Number of rows of data
     */
   virtual unsigned long numRows() = 0;
   /** Function to be implemented in the derived class for getting the number
     * of rows left in the result set for a given query
     * \return Number of rows left that can be accessed
     */
   virtual unsigned long rowsLeft() = 0;

   /** Sets the given flag
     * \param _flag Flag to set
     */
   void setFlag(unsigned int _flag)
   {
      dataFlag |= (1 << _flag);
   }

   /** Checks whether the given flag is set or unset
     * \param _flag Flag to check
     * \return Returns 0 if not set else greater than 0
     */
   unsigned int isFlagOn(unsigned int _flag)
   {
      return ( dataFlag & (1 << _flag) );
   }

   /** Clear the given flag
     * \param _flag Flag to clear
     */
   void clearFlag(unsigned int _flag)
   {
      dataFlag &= ~(1 << _flag);
   }

public:
   /** Row data is stored as a pointer. Should not be destroyed when deleting
     * this object */
   char **rowData;
   /** Port number for the connection */
   unsigned int portNumber;
   /** Flag for knowing the state of reading data */
   unsigned int dataFlag;
   /** Number of rows of data in the result set after a query */
   unsigned long numberOfRows;
   string dbVendor, dbName, hostName, userName, passWord;
};

#endif

