#ifndef _MY_SQL_CONN_H_
#define _MY_SQL_CONN_H_

#include <mysql/mysql.h>
#include "goldenconn.h"

/** Class to access MySQL database. For each vendor, this type of object is
  * created to access the database.
  */
class MySqlConn : public GoldenConn
{
public:
   /** Constructor to create the object
     * \param db_name Name of database
     * \param host Name of the host or IP address of the database machine
     * \param user Name of the user
     * \param passwd Password for user
     * \param port_num Port number to the database
     */
   MySqlConn
   (
      const char  *db_name,
      const char  *host,
      const char  *user,
      const char  *passwd,
      unsigned int port_num = 3306
   ) : GoldenConn(db_name, host, user, passwd, port_num)
   {
      myConnection = NULL;
      myResult = NULL;
   }

   ~MySqlConn()
   {
      if( myConnection )
         mysql_close( myConnection );
      myConnection = NULL;

      if( myResult )
         mysql_free_result( myResult ); 
   }

   /** Connects to MySQL database
     * \return Returns 0 on success else greater than 0
     */
   int connect()
   {
      if( myConnection )
         return 0;

      myConnection = NULL;
      myConnection = mysql_init(myConnection);

      if( !myConnection )
	      return 1;

      if( !mysql_real_connect(myConnection, hostName.data(), userName.data(),
	   passWord.data(), dbName.data(), portNumber, NULL, 0) )
      {
         cout << mysql_error( myConnection ) << endl;
         return 2;
      }

      return 0;
   }

   /** Executes the given query and gets the results from MySQL database
     * \param qry Query to be executed on MySQL
     * \param store_result true - to store the results in this object or false -
     * not to store the results in this object
     * \return Returns 0 on success else greater than 0
     */
   int executeQuery(const char *qry, bool store_result = true)
   {
      if( myResult )
         mysql_free_result( myResult );
      numberOfRows = 0;
      dataFlag = 0;
      myResult = NULL;

      if( connect() )
         return 3;

      if( mysql_query(myConnection, qry) )
      {
	 cout << mysql_error( myConnection ) << endl;
	 return 4;
      }

      if( store_result )
      {
         myResult = mysql_store_result( myConnection );
         if( myResult )
            numberOfRows = mysql_num_rows( myResult );
      }
      return 0;
   }

   /** Gets the number of rows still remaining after traversing the result
     * line by line
     * \return Number of rows still remaining in the result
     */
   unsigned long rowsLeft()
   {
      return numberOfRows > 0 ? numberOfRows : 0;
   }

   /** Gets the count of rows after executing the query
     * \return Number of rows in the result
     */
   unsigned long numRows()
   {
      return mysql_num_rows( myResult );
   }

   /** Gets the next row of data from the result
     * \return Returns entire row of data 
     */
   char **nextRow()
   {
      rowData = mysql_fetch_row( myResult );

      if( rowData )
      {
         numberOfRows--;
         return rowData;
      }

      numberOfRows = 0L;
      return NULL;
   }

   /** Gets the next card from the result
     * \return Returns entire card data 
     */
   char **nextCard()
   {
      if( isFlagOn(KW_DATA_END) )
         return NULL;
          
      rowData = mysql_fetch_row( myResult );

      if( rowData && rowData[2] && rowData[2][0] == '$' )
      {
         setFlag( KW_DATA_END );
         numberOfRows--;
         return NULL;
      }

      if( rowData )
      {
         numberOfRows--;
         return rowData;
      }

      numberOfRows = 0L;
      return NULL;
   }

   /** Gets the error message if any on the server after executing the query */
   const char *error()
   {
      return mysql_error(myConnection);
   }

   /** Sets the given MySQL option to the server while getting the data
     * \param opt Option to set to the MySQL MYSQL_OPTION_MULTI_STATEMENTS_ON,
     * MYSQL_OPTION_MULTI_STATEMENTS_OFF
     * \return Returns true on success else false
     */
   int setOption(unsigned int opt)
   {
      return mysql_set_server_option(myConnection, (enum_mysql_set_option)opt);
   }

private:
   MYSQL *myConnection;
   MYSQL_RES *myResult;
};

#endif

