#ifndef _FILE_PATH_H_
#define _FILE_PATH_H_

#include <iostream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "bufio.h"
#include "transform.h"

/** Class to save the node data, element connectivity, entity IDS, referred 
  * entity IDs to files while getting the data from database simultaneously 
  * without storing any data from database.
  */
class FilePath
{
public:
   /** Sets the directory and name of file to save
     * \param dir Name of directory to where the file is saved
     * \param model Name of model which is also the file name
     */
   static void set(const char *dir, const char *model)
   {
      strcpy(dirName, dir);
      strcpy(modelName, model);
      sprintf(fPath, "%s/%s", dir, model);
      if( access(fPath, W_OK) )
         mkdir(fPath, 0755);
   }

   static void setPath(const char *dir, const char *model, int dirIndex)
   {
      strcpy(dirName, dir);
      strcpy(modelName, model);
      sprintf(fPath, "%s/%s/%d", dir, model, dirIndex);
      if( access(fPath, W_OK) )
         mkdir(fPath, 0755);
   }

   /** Sets the directory, name of file to save and also a transform pointer to
     * transform the model. The TransformPtr contains the infromation of
     * *DEFINE_TRANSFORMATION keyword. ID offsets and transformations are applied
     * before saving the data to files.
     * \param dir Name of directory to where the file is saved
     * \param model Name of model which is also the file name
     * \param trans_ptr Transformation details
     */
   static void set(const char *dir, const char *model, TransformPtr trans_ptr)
   {
      transPtr = trans_ptr;
      set(dir, model);
   }

   /** Sets a transform pointer to transform the model. The TransformPtr contains 
     * the infromation of *DEFINE_TRANSFORMATION keyword. ID offsets and 
     * transformations are applied before saving the data to files.
     * \param trans_ptr Pointer to Transformation details
     */
   static void setTransPtr(TransformPtr trans_ptr)
   {
      transPtr = trans_ptr;
   }

   /** Gets the directory and name of model 
     * \param dir Name of directory
     * \param dir Name of model
     */
   static void get(char **dir, char **model)
   {
      *dir = dirName;
      *model = modelName;
   }

   /** Opens an index file. The index file stores the information of card_index
     * in the database and the element IDS and the part ID. Based on the card 
     * index, the entity ID or part ID can be found.
     * \param file Name of file to which the indices are to be saved
     * \return Returns 0 on SUCCESS else greater than 0
     */
   static int openIndexFile(const char *file)
   {
      indexFileDes = open(file, O_RDWR|O_APPEND|O_CREAT,
			   S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
      if( indexFileDes < 0 )
      {
         indexFileDes = -2;
	 cout << "Unable to open file for writing in filepath.h\n";
	 return 1;
      }

      indexBufPtr = BufFileOpenWrite(indexFileDes);

      return 0;
   }

   /** Opens a file for storing the information of node values and element
     * connectivity for rendering the model. Each file has the part id and 
     * elements connectivity.
     * \param file Name of file to which the data to be saved
     * \return Returns 0 on SUCCESS else greater than 0
     */
   static int openFile(const char *file)
   {
      fileDes = open(file, O_RDWR|O_APPEND|O_CREAT,
			   S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
      if( fileDes < 0 )
      {
         fileDes = -2;
	 cout << "Unable to open file for writing in filepath.h\n";
	 return 1;
      }

      bufPtr = BufFileOpenWrite(fileDes);

      return 0;
   }

   /** Opens a file for storing the information of card indices, entity ID and
     * keyowrd type. The keyword type information is used to display the
     * referred entities.
     * \param file Name of file to which the data to be saved
     * \param flag Overwrite or append
     * \return Returns 0 on SUCCESS else greater than 0
     */
   static int openIDFile(const char *file, int flag = O_TRUNC)
   {
      idFileDes = open(file, O_RDWR|flag|O_CREAT,
			   S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH);
      if( idFileDes < 0 )
      {
         idFileDes = -2;
	 cout << "Unable to open file for writing in openIDFile " <<file<<endl;
	 return 1;
      }

      idBufPtr = BufFileOpenWrite(idFileDes);
cout << "FILE OPENED ID FILE" << endl;

      return 0;
   }

   /** Closes the index file already opened */
   static void closeIndexFile()
   {
      if( indexBufPtr )
      {
	 BufFileClose(indexBufPtr, 1);
         indexBufPtr = NULL;
	 close(indexFileDes);
         indexFileDes = -1;
      }
   }

   /** Closes the data file already opened */
   static void closeFile()
   {
      if( bufPtr )
      {
	 BufFileClose(bufPtr, 1);
         bufPtr = NULL;
	 close(fileDes);
         fileDes = -1;
      }
   }

   /** Closes the ID file already opened */
   static void closeIDFile()
   {
      if( idBufPtr )
      {
	 BufFileClose(idBufPtr, 1);
         idBufPtr = NULL;
	 close(idFileDes);
         idFileDes = -1;
cout << "FILE CLOSED ID FILE" << endl;
      }
   }

public:
   static char dirName[128], modelName[128], fPath[256];
   /** Current part ID */
   static unsigned int pID;
   /** File descriptor for the elements binary file */
   static int fileDes;
   /** File descriptor for the IDs binary file */
   static int idFileDes;
   /** File descriptor for elements card index, element id and part id */
   static int indexFileDes;
   /** File pointer for the elements binary file */
   static BufFilePtr bufPtr;
   /** File pointer for the IDs binary file */
   static BufFilePtr idBufPtr;
   /** File pointer for elements card index, element id and part id */
   static BufFilePtr indexBufPtr;
   /** Transform details pointer */
   static TransformPtr transPtr;
};

#endif

