#include <iostream>
#include <algorithm>
#include "generic.h"
#include "bdrytype.h"
#include "bdryimport.h"
#include <stdlib.h>

using namespace std;

BdryImport::BdryImport
(
   FILE          *fp,
   DataLoc       *data_loc,
   BdryManager    *bdry_man
)
{
   filePtr = fp;
   bdryManPtr = bdry_man;
   dataLocPtr = data_loc;
   dataLine = NULL;

   initSimpleTypesAndFunctions();
   initBEMTypesAndFunctions();
   initOrientTypesAndFunctions();
   initPrescribedTypesAndFunctions();
   initRadiationTypesAndFunctions();
   initSpcTypesAndFunctions();
   initTemperatureTypesAndFunctions();
   initOutflowTypesAndFunctions();
   initConvectionTypesAndFunctions();
   initFluxTypesAndFunctions();
}

BdryType BdryImport::getBdryTypeFromKeyword
	(
	   const char  *key_word,
	   unsigned int *bdry_id
	)
{
   *bdry_id = 0;
   if( strstr(key_word, "IC_CO") )
      return BOUNDARY_ACOUSTIC_COUPLING;
   else if( strstr(key_word, "NT_EO") )
      return BOUNDARY_AMBIENT_EOS;
   else if( strstr(key_word, "RY_CON") )
   {
      if( strstr(key_word, "_SET") )
         return BOUNDARY_CONVECTION_SET;
      else
         return BOUNDARY_CONVECTION_SEGMENT;
   }
   else if( strstr(key_word, "RY_CY") )
      return BOUNDARY_CYCLIC;
   else if( strstr(key_word, "UX_SEG") )
      return BOUNDARY_FLUX_SEGMENT;
   else if( strstr(key_word, "UX_SET") )
      return BOUNDARY_FLUX_SET;
   else if( strstr(key_word, "_MCOL") )
      return BOUNDARY_MCOL;
   else if( strstr(key_word, "ON_REF") )
   {
      if( strstr(key_word, "_2D") )
         return BOUNDARY_NON_REFLECTING_2D;
      else
         return BOUNDARY_NON_REFLECTING;
   }
   else if( strstr(key_word, "ER_RI") )
      return BOUNDARY_PRESCRIBED_ACCELEROMETER_RIGID;
   else if( strstr(key_word, "NG_PL") )
      return BOUNDARY_SLIDING_PLANE;
   else if( strstr(key_word, "PH_FL") )
      return BOUNDARY_SPH_FLOW;
   else if( strstr(key_word, "RY_PL") )
      return BOUNDARY_SPH_SYMMETRY_PLANE;
   else if( strstr(key_word, "RY_FA") )
      return BOUNDARY_SYMMETRY_FAILURE;
   else if( strstr(key_word, "RE_NO") )
      return BOUNDARY_TEMPERATURE_NODE;
   else if( strstr(key_word, "RE_SE") )
      return BOUNDARY_TEMPERATURE_SET;
   else if( strstr(key_word, "AL_WE") )
      return BOUNDARY_THERMAL_WELD;
   else if( strstr(key_word, "SA_SU") )
      return BOUNDARY_USA_SURFACE;
   else if( strstr(key_word, "OW_SEG") )
      return BOUNDARY_PRESSURE_OUTFLOW_SEGMENT;
   else if( strstr(key_word, "OW_SET") )
      return BOUNDARY_PRESSURE_OUTFLOW_SET;
   else if( strstr(key_word, "OD_AC") )
      return BOUNDARY_ELEMENT_METHOD_ACOUSTIC;
   else if( strstr(key_word, "OD_CON") )
      return BOUNDARY_ELEMENT_METHOD_CONTROL;
   else if( strstr(key_word, "OD_FL") )
      return BOUNDARY_ELEMENT_METHOD_FLOW;
   else if( strstr(key_word, "OD_SY") )
      return BOUNDARY_ELEMENT_METHOD_SYMMETRY;
   else if( strstr(key_word, "OD_NE") )
      return BOUNDARY_ELEMENT_METHOD_NEIGHBOR;
   else if( strstr(key_word, "OD_WA") )
      return BOUNDARY_ELEMENT_METHOD_WAKE;

   else if( strstr(key_word, "ID_AN") )
      return BOUNDARY_PRESCRIBED_ORIENTATION_RIGID_ANGLES;
   else if( strstr(key_word, "ID_DI") )
      return BOUNDARY_PRESCRIBED_ORIENTATION_RIGID_DIRCOS;
   else if( strstr(key_word, "ID_EU") )
      return BOUNDARY_PRESCRIBED_ORIENTATION_RIGID_EULERP;
   else if( strstr(key_word, "_RAD") )
   {
      if( strstr(key_word, "NT_VF_R") )
         return BOUNDARY_RADIATION_SEGMENT_VF_READ;
      else if( strstr(key_word, "NT_VF_C") )
         return BOUNDARY_RADIATION_SEGMENT_VF_CALCULATE;
      else if( strstr(key_word, "N_SEG") )
         return BOUNDARY_RADIATION_SEGMENT;
      else if( strstr(key_word, "T_EF_R") )
         return BOUNDARY_RADIATION_SET_EF_READ;
      else if( strstr(key_word, "T_EF_C") )
         return BOUNDARY_RADIATION_SET_EF_CALCULATE;
      else if( strstr(key_word, "ET_VF_R") )
         return BOUNDARY_RADIATION_SET_VF_READ;
      else if( strstr(key_word, "ET_VF_C") )
         return BOUNDARY_RADIATION_SET_VF_CALCULATE;
      else
         return BOUNDARY_RADIATION_SET;
   }
   else if( strstr(key_word, "_MOT") )
   {
      if( strstr(key_word, "_ID") )
      {
         if( strstr(key_word, "N_NO") )
            return BOUNDARY_PRESCRIBED_MOTION_NODE_ID;
         else if( strstr(key_word, "D_I") )
            return BOUNDARY_PRESCRIBED_MOTION_RIGID_ID;
         else if( strstr(key_word, "D_L") )
            return BOUNDARY_PRESCRIBED_MOTION_RIGID_LOCAL_ID;
         else if( strstr(key_word, "N_S") )
            return BOUNDARY_PRESCRIBED_MOTION_SET_ID;
      }
      else
      {
         if( strstr(key_word, "N_NO") )
            return BOUNDARY_PRESCRIBED_MOTION_NODE;
         else if( strstr(key_word, "D_L") )
            return BOUNDARY_PRESCRIBED_MOTION_RIGID_LOCAL;
         else if( strstr(key_word, "N_S") )
            return BOUNDARY_PRESCRIBED_MOTION_SET;
         else
            return BOUNDARY_PRESCRIBED_MOTION_RIGID;
      }
   }
   else if( strstr(key_word, "_SPC") )
   {
      if( strstr(key_word, "_ID") )
      {
         if( strstr(key_word, "C_N") )
            return BOUNDARY_SPC_NODE_ID;
         else if( strstr(key_word, "C_S") )
            return BOUNDARY_SPC_SET_ID;
      }
      else
      {
         if( strstr(key_word, "C_N") )
            return BOUNDARY_SPC_NODE;
         else if( strstr(key_word, "C_S") )
            return BOUNDARY_SPC_SET;
      }
   }

   return BDRY_NEW_TYPE;
}

unsigned int BdryImport::readBdrys()
{
   char         new_kw[LINE_LEN];
   unsigned int	i = 0, bdry_id = 0, bdry_type = 0, num_bdrys = 0;
   long		loc = 0L;
   VStringIterator   s = dataLocPtr->keywordList.begin(),
		     e = dataLocPtr->keywordList.end(), itr, 
		     pitr = dataLocPtr->keywordList.begin();
   typeFuncIterator  btype_iter = typeFuncMap.begin(),
                     etype_iter = typeFuncMap.end(), type_iter;

   dataLine = new DataLine(filePtr);

   while(true)
   {
      itr = find_if(pitr, e, bind2nd(KeyWordCompare (), "*BOUN"));

      if( itr == e )
	 break;

      pitr = itr+1;
      i = (unsigned int)(itr - s);
      loc = dataLocPtr->slocList[i];

      rewind(filePtr);

      if( fseek(filePtr, loc, SEEK_SET) )
      {
	 cout << "ERROR in locating the Boundary cards" << endl;
	 break;
      }

      dataLine->clearFlag(KW_READ_END);

      char *str = dataLine->readNextCard(); //get the keyword
      Generic::RemoveLeadingAndTrailingBlanks( str );
cout << "KW = " << str << endl;

      bdry_type = getBdryTypeFromKeyword( str, &bdry_id );
cout << "Bdry TYPE = " << bdry_type << endl;

      if( bdry_type == BDRY_NEW_TYPE )
      {
         cout << "Boundary TYPE NOT SUPPORTED" << endl;
         continue;
      }

      num_bdrys++;

      strcpy(new_kw, str);

      dataLine->clearFlag(KW_READ_END);
      type_iter = typeFuncMap.find(bdry_type);
      if( type_iter != etype_iter )
      {
          impObj = new ImportObject(str, bdry_type);
          (type_iter->second)(dataLine, impObj);
          if( bdry_id )
          {
             bdryManPtr->add(new_kw, impObj);
          }
          else
          {
             sprintf(new_kw, "%s_%d", new_kw, num_bdrys);
             bdryManPtr->add(new_kw, impObj);
          }
      }
      else
         cout << "Boundary TYPE NOT FOUND" << endl;
   }

   return num_bdrys;
}

