#ifndef _Contact_IMPORT_H_
#define _Contact_IMPORT_H_

#include <string>
#include <map>
#include "dataloc.h"
#include "contactmanager.h"
#include "dataline.h"
#include "card.h"
#include "contactimpobj.h"
#include "contacttype.h"
#include "typemap.h"

/** Class to import the data from native LSDYNA file for *CONTACT keywords. All
  * the data is imported through this class and stored in the manager object.
  */
class ContactImport
{
public:
   /** Constructor for creating this object
     * \param fp File pointer to the LSDYNA native file
     * \param data_loc Pointer to the list of keywords and their locations in
     * the file
     * \param contact_man Manager pointer to hold the imported data as a list of
     * ImportObject list
     */
   ContactImport
   (
      FILE          *fp,
      DataLoc       *data_loc,
      ContactManager    *contact_man
   );

   ~ContactImport()
   {
      filePtr = NULL;
      contactManPtr = NULL;
      dataLocPtr = NULL;
      delete dataLine;
      typeFuncMap.clear();
   }

   /** Function for getting the keyword type from given keyword name from LSDYNA
     * native file. The type is stored in the database
     * \param key_word Name of the keyword from file
     * \return Type for *CONTACT_GEBOD group of keywords
     */
   static ContactType getGebodTypeFromKeyword
	       (
		  const char *key_word
	       );

   /** Function for getting the keyword type from given keyword name from LSDYNA
     * native file. The type is stored in the database
     * \param key_word Name of the keyword from file
     * \return Type for *CONTACT_2D group of keywords
     */
   static ContactType get2DContactTypeFromKeyword
	       (
		  const char *key_word
	       );

   /** Function for getting the keyword type from given keyword name from LSDYNA
     * native file. The type is stored in the database
     * \param key_word Name of the keyword from file
     * \param contact_bin A flag to identify the type of *CONTACT
     * \return Type for *CONTACT group of keywords
     */
   static ContactType getContactTypeFromKeyword
	       (
		  const char *key_word,
                  unsigned int *contact_bin
	       );

   /** Reads the *CONTACT keywords from file
     * \return Returns 0 on failure else number of *CONTACT keywords
     */
   unsigned int readContacts();

   /** Creates keyword type and pointer to function initializations. This
     * function is a generated code from the XML files for keywords
     * For each type of *CONTACT keyword there is a function to call
     */
   void initSimpleTypesAndFunctions();

private:
   FILE              *filePtr;
   /** Manager pointer for holding the imported data for all the keywords */
   ContactManager    *contactManPtr;
   /** Pointer to keywords and their locations in the file */
   DataLoc           *dataLocPtr;
   /** Pointer to access the file data line by line */
   DataLine          *dataLine;
   /** Object to store imported data from LSDYNA data file */
   ImportObject      *impObj;
   /** A map between the keyword type and pointer to function */
   map<unsigned int, ImportFunctionPtr>  typeFuncMap;
};

#endif

