#ifndef _CARD_H_
#define _CARD_H_

#include "object.h"
#include "datatypeobj.h"
#include "bufio.h"
#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include <unistd.h>

#define SAVE_FIELD_NAMES (1<<1)
#define SAVE_CARDS_FIELDS (1<<2)
#define SAVE_IDS (1<<3)
#define SAVE_REFERENCES (1<<4)

#define IMPORT_DYNA (1<<10)
#define IMPORT_DB (1<<11)
#define EXPORT_DYNA (1<<12)
#define EXPORT_DB (1<<13)

using namespace std;

/** This class represents a line of data or a card in LSDYNA format. Each
  * card has specified number of fields. All fields have values. The fields
  * with their names and values are stored here. The field value can be of 
  * type int, double or a string. The value is stored as a generic data object.
  */ 

class Card : public Object
{
public:
   /** Constructor to create a card object from the name of it.
     * \param name Name of the card to create
     */
   Card(const char *name)
   {
      cardName = name;
   }

   /** Constructor to create a card object from the name of it and a flag to
     * indicate whether the name of the field has to be stored each time 
     * or not. For example, *NODE has values x,y,z,tx,ty,tz. The name for
     * one card is enough to store but all the values are to be saved. Using
     * the _flag, one can specify to store the repeated names or not.
     * \param name Name of the card to create
     * \param _flag Flag to save the repeated names for the card or not
     */
   Card(const char *name, bool _flag)
   {
      cardName = name;
      if( _flag )
         clearFlag();
   }

   /** Destructor to release the memory in the object */
   virtual ~Card()
   {
      vector<Data *>::iterator b = varValues.begin(), e = varValues.end();
      //Data *data;
#ifdef DEBUG
cout << "CARD NAME=" << getName() << endl;
#endif

      for(; b != e; b++)
      {
	 //data = *b;
         //if( data )
	    delete (*b);
      }

      varValues.clear();
      varNames.clear();
   }

   /** Function to add a list of double type fields to this object 
     * \param num_fields Number of fields to add
     * \param field_types Type for each field.
     * \param field_names List of field names to add
     * \param dbl_values List of double type values
     */
   void addFields
        (
           unsigned int   & num_fields,
           unsigned int   field_types[], 
           const char     *field_names[],
           double         *dbl_values
        )
   {
      unsigned int uint_val;
      int          int_val;
      UInteger   *uint_obj;
      Integer    *int_obj;
      Double     *dbl_obj;

      for(unsigned int i = 0; i < num_fields; i++)
      {
         switch(field_types[i])
         {
            case DATA_UINT:
               uint_val = (unsigned int)dbl_values[i];
               uint_obj = new UInteger(uint_val);
               addField(field_names[i], uint_obj); 
               break;
            case DATA_INT:
               int_val = (int)dbl_values[i];
               int_obj = new Integer(int_val);
               addField(field_names[i], int_obj); 
               break;
            case DATA_DOUBLE:
               dbl_obj = new Double(dbl_values[i]);
               addField(field_names[i], dbl_obj);
               break;
            case DATA_UNUSED:
               addEmptyField();
               break;
         }
      } 
   }

   /** Function to add a list of string type fields having different width 
     * to this object 
     * \param num_fields Number of fields to add
     * \param field_types Type for each field.
     * \param field_widths List of fields width
     * \param field_names List of field names to add
     * \param word_array List of string type values
     */
   void addFields
        (
           unsigned int   & num_fields,
           unsigned int   field_types[], 
           unsigned int   field_widths[], 
           const char     *field_names[],
           char           **word_array
        )
   {
      UInteger   *uint_obj;
      Integer    *int_obj;
      Double     *dbl_obj;
      CharArray  *str_obj;
      //BlankObject *bo;

      for(unsigned int i = 0; i < num_fields; i++)
      {
         if( word_array[i] && !iscntrl(word_array[i][0]) && word_array[i][0] != '\0' && word_array[i][0] != '\n' )
         {
	    switch(field_types[i])
	    {
	       case DATA_UINT:
		  uint_obj = new UInteger(word_array[i], field_widths[i]);
		  addField(field_names[i], uint_obj); 
		  break;
	       case DATA_INT:
		  int_obj = new Integer(word_array[i], field_widths[i]);
		  addField(field_names[i], int_obj); 
		  break;
	       case DATA_DOUBLE:
		  dbl_obj = new Double(word_array[i], field_widths[i]);
		  addField(field_names[i], dbl_obj);
		  break;
	       case DATA_CHAR_ARRAY:
		  str_obj = new CharArray(word_array[i], field_widths[i]);
		  addField(field_names[i], str_obj);
		  break;
	       case DATA_UNUSED:
		  addEmptyField();
		  break;
	    }
         }
         else
         {
            //adding these blank fields increases the file size
            //bo = new BlankObject(field_widths[i]);
	    //addField(field_names[i], bo);
            blankFields.push_back(field_names[i]);
         }
      }
   }

   /** Function to add a list of string type fields to this object 
     * \param num_fields Number of fields to add
     * \param field_types Type for each field.
     * \param field_names List of field names to add
     * \param word_array List of string type values
     */
   void addFields
        (
           unsigned int   & num_fields,
           unsigned int   field_types[], 
           const char     *field_names[],
           char           **word_array
        )
   {
      unsigned int uint_val;
      int          int_val;
      UInteger   *uint_obj;
      Integer    *int_obj;
      Double     *dbl_obj;
      CharArray  *str_obj;
      double     dbl_val;
      //BlankObject  *bo;

      for(unsigned int i = 0; i < num_fields; i++)
      {
         if( word_array[i] && !iscntrl(word_array[i][0]) && 
                         word_array[i][0] != '\0' && word_array[i][0] != '\n' )
         {
	    switch(field_types[i])
	    {
	       case DATA_UINT:
		  uint_val = atoi(word_array[i]);
		  uint_obj = new UInteger(uint_val);
		  addField(field_names[i], uint_obj); 
		  break;
	       case DATA_INT:
		  int_val = atoi(word_array[i]);
		  int_obj = new Integer(int_val);
		  addField(field_names[i], int_obj); 
		  break;
	       case DATA_DOUBLE:
		  dbl_val = atof(word_array[i]);
		  dbl_obj = new Double(dbl_val);
		  addField(field_names[i], dbl_obj);
		  break;
	       case DATA_CHAR_ARRAY:
		  str_obj = new CharArray(word_array[i]);
		  addField(field_names[i], str_obj);
		  break;
	       case DATA_UNUSED:
		  addEmptyField();
		  break;
	    }
         }
         else
         {
            //adding these blank fields increases the file size
            //bo = new BlankObject();
	    //addField(field_names[i], bo);
            blankFields.push_back(field_names[i]);
         }
      }
   }

   /** Function to add a list of fields values to this object 
     * \param num_fields Number of fields to add
     * \param field_types Type for each field.
     * \param field_widths List of fields width
     * \param field_names List of field names to add
     * \param str_buff Raw buffer for entire card having all field values
     */
   void addFields
        (
           unsigned int   & num_fields,
           unsigned int   field_types[],
           unsigned int   field_widths[],
           const char     *field_names[],
           char           *str_buff
        )
   {
      char         str_word[128];
      unsigned int idx = 0, i; //uint_val, idx = 0;
      //int          int_val;
      Integer    *int_obj;
      UInteger   *uint_obj;
      Double     *dbl_obj;
      CharArray  *str_obj;
      //double     dbl_val;
      //BlankObject  *bo;

      if( !strchr(str_buff, ',') )
      {
	 for(i = 0; i < num_fields; i++)
	 {
	    memset(str_word, '\0', 128);
	    memcpy(str_word, &str_buff[idx], field_widths[i]); 
	    idx += field_widths[i];

	    if( !iscntrl(str_word[0]) && str_word[0] != '\0' && 
                                                      str_word[0] != '\n' )
	    {
	       switch(field_types[i])
	       {
		  case DATA_UINT:
		     uint_obj = new UInteger(str_word, field_widths[i]);
		     addField(field_names[i], uint_obj); 
		     break;
		  case DATA_INT:
		     int_obj = new Integer(str_word, field_widths[i]);
		     addField(field_names[i], int_obj); 
		     break;
		  case DATA_DOUBLE:
		     dbl_obj = new Double(str_word, field_widths[i]);
		     addField(field_names[i], dbl_obj);
		     break;
		  case DATA_CHAR_ARRAY:
		     str_obj = new CharArray(str_word, field_widths[i]);
		     addField(field_names[i], str_obj);
		     break;
		  case DATA_UNUSED:
		     addEmptyField(1, field_widths[i]);
		     break;
	       }
	    }
	    else
	    {
               //adding blank field increasing the file size
               //bo = new BlankObject(field_widths[i]);
	       //addField(field_names[i], bo);
               blankFields.push_back(field_names[i]);
	    }
	 }
      }
      else
      {
         i = 0;
         char *str_char = strtok(str_buff, ",");

	 while( str_char )
	 {
            if(i >= num_fields)
               break;

	    switch(field_types[i])
	    {
	       case DATA_UINT:
		  uint_obj = new UInteger(str_char, field_widths[i]);
		  addField(field_names[i], uint_obj); 
		  break;
	       case DATA_INT:
		  int_obj = new Integer(str_char, field_widths[i]);
		  addField(field_names[i], int_obj); 
		  break;
	       case DATA_DOUBLE:
		  dbl_obj = new Double(str_char, field_widths[i]);
		  addField(field_names[i], dbl_obj);
		  break;
	       case DATA_CHAR_ARRAY:
		  str_obj = new CharArray(str_char, field_widths[i]);
		  addField(field_names[i], str_obj);
		  break;
	       case DATA_UNUSED:
		  addEmptyField(1, field_widths[i]);
		  break;
	    }
            str_char = strtok(NULL, ",");
            i++;
	 }
      }
   }

   /** Function to add a field value to this object 
     * \param name field name to add
     * \param val Data object holding the value of the field
     */
   void addField(const char *name, Data *val)
   {
      if( contFlag & (1 << 1 ) )
         varNames.push_back(name);
      varValues.push_back(val);
#ifdef DEBUG
cout << name;
      unsigned int uint_val = 0;
      int int_val = 0;
      double dbl_val = 0;

      switch(val->dataType)
      {
         case DATA_UINT:
            val->value(&uint_val);
cout << "=" << uint_val << endl;
            break;
         case DATA_INT:
            val->value(&int_val);
cout << "=" << int_val << endl;
            break;
         case DATA_DOUBLE:
            val->value(&dbl_val);
cout << "=" << dbl_val << endl;
            break;
      }
#endif
   }

   /** Function to add an empty field value to this object 
     * \param num Num of empty fields to add. Default is 1.
     * \param width Width of field to add. Default is 10.
     */
   void addEmptyField(int num = 1, unsigned int width = 10)
   {
      static char *unused = (char *)" ";
      int         i;
      
      for(i = 0; i < num; i++)
      {
	 if( contFlag & (1 << 1 ) )
	    varNames.push_back(unused);
         BlankObject *bo = new BlankObject(width);
	 varValues.push_back(bo);
      }
   }

   /** Function to write this card as a line of data to the given file pointer.
     * \param buf_ptr Pointer to the file as a BufFilePtr
     */
   void toFile(BufFilePtr buf_ptr)
   {
      char          buff[128], format_str[16];
      Data          *data;

      vector<Data *>::iterator bv = varValues.begin(),ev = varValues.end(),vit;
      vector<string>::iterator b = varNames.begin(), e = varNames.end(), iter;

      if( varNames.size() )
      {
	 data = (*bv);
	 sprintf(format_str, "$%%%us", (data->dataWidth)-1);

	 sprintf(buff, format_str, (*b).data());
	 b++; 
	 bv++;
	 BufFileWrite(buf_ptr, buff, strlen(buff));
	 for(iter = b; iter < e; iter++)
	 {
	    data = *bv;
	    sprintf(format_str, "%%%us", data->dataWidth);
	    sprintf(buff, format_str, (*iter).data());
	    BufFileWrite(buf_ptr, buff, strlen(buff));
	    bv++;
	 }
	 BufFileWrite(buf_ptr, (char *)"\n", 1);
      }

      bv = varValues.begin(); 
      ev = varValues.end();

      for(vit = bv; vit < ev; vit++)
      {
         data = *vit;
         data->tostring(buff);
         BufFileWrite(buf_ptr, buff, strlen(buff));
      }
      if( buff[strlen(buff)-1] != '\n')
	 BufFileWrite(buf_ptr, (char *)"\n", 1);
   }

   /** Function to covert this object as a line of data to a string value.
     * \param str String to which this object is to be added
     */
   void tostring(string & str)
   {
      char          buff[128], format_str[16];
      Data          *data;

      vector<Data *>::iterator bv = varValues.begin(),ev = varValues.end(),vit;
      vector<string>::iterator b = varNames.begin(), e = varNames.end(), iter;

      if( varNames.size() )
      {
	 data = (*bv);
	 sprintf(format_str, "$%%%us", (data->dataWidth)-1);

	 sprintf(buff, format_str, (*b).data());
	 b++; 
	 bv++;
	 str += buff;
	 for(iter = b; iter < e; iter++)
	 {
	    data = *bv;
	    sprintf(format_str, "%%%us", data->dataWidth);
	    sprintf(buff, format_str, (*iter).data());
	    str += buff;
	    bv++;
	 }
	 str += "\n";
      }

      bv = varValues.begin(); 
      ev = varValues.end();

      for(vit = bv; vit < ev; vit++)
      {
         data = *vit;
         data->tostring(buff);
         str += buff;
      }
      if( buff[strlen(buff)-1] != '\n')
         str += "\n";
   }

   /** No functionality */
   void tostring(char *)
   {
   }

   /** Function to get the unsigned int value of a variable
     * \param var_name Name of variable whose type is an unsigned int
     * \return Returns unsigned int value if found otherwise zero.
     */
   unsigned int uintValue
		(
		   const char  *var_name
		)
   {
      unsigned int val = 0;

      vector<string>::iterator  b = varNames.begin(), e = varNames.end(), iter;
      iter = find(b, e, var_name);
      if( iter != e )
         varValues[iter-b]->value(&val);

      return val;
   }

   /** Function to get the int value of a variable
     * \param var_name Name of variable whose type is an int
     * \return Returns int value if found otherwise zero.
     */
   int  intValue
        (
           const char  *var_name
        )
   {
      int val = 0;

      vector<string>::iterator  b = varNames.begin(), e = varNames.end(), iter;
      iter = find(b, e, var_name);
      if( iter != e )
         varValues[iter-b]->value(&val);
      return val;
   }

   /** Function to get value of a variable whose type is a double
     * \param var_name Name of variable whose type is a double
     * \return Returns double value if found otherwise zero.
     */
   double doubleValue
	  (
	     const char  *var_name
	  )
   {
      double val = 0;

      vector<string>::iterator  b = varNames.begin(), e = varNames.end(), iter;
      iter = find(b, e, var_name);
      if( iter != e )
         varValues[iter-b]->value(&val);
      return val;
   }

   /** Function to get value of a variable whose type is string
     * \param var_name Name of variable whose type is a string
     * \return Returns string value if found otherwise empty.
     */
   char *stringValue
	(
	   const char  *var_name
	)
   {
      char *val = (char *)"";
      vector<string>::iterator  b = varNames.begin(), 
                                e = varNames.end(), iter;
      iter = find(b, e, var_name);
      if( iter != e )
         varValues[iter-b]->value(&val);
      return val;
   }

   /** Function to find a field with a given name
     * \param field_name Name of the field
     * \return Returns Data object holding the value of variable.
     */
   Data *findField
         (
            const char  *field_name
         )
   {
      vector<string>::iterator b = varNames.begin(), e = varNames.end(), iter;
      iter = find(b, e, field_name);

      if( iter != e )
         return varValues[iter-b];
      
      return NULL;
   }

   /** Get the name of this object as a reference
     * \return The name of card object
     */
   string & getName()
   {
      return cardName;
   }

   /** Set a given flag to this object
     * \param flag Value of flag. Default is SAVE_FIELD_NAMES
     */
   static void setFlag(unsigned int flag = SAVE_FIELD_NAMES)
   {
      contFlag |= flag;
   } 

   /** Clear a given flag to this object
     * \param flag Value of flag to clear. Default is SAVE_FIELD_NAMES
     */
   static void clearFlag(unsigned int flag = SAVE_FIELD_NAMES)
   {
      contFlag &= ~flag;
   } 

   /** Checks whether the given flag is set or not
     * \param flag Value of flag. Default is SAVE_FIELD_NAMES
     * \return Returns zero if not set otherwise an unsigned int
     */
   static unsigned int isFlagOn(unsigned int flag = SAVE_FIELD_NAMES)
   {
      return (contFlag & flag);
   }

   /** Gets the reference to the list of variable names
     * \return Returns a reference to the array of field names
     */
   vector<string> & getVarNames()
   {
      return varNames;
   }

   /** Gets the reference to the list of variable values
     * \return Returns a reference to the array of field values
     */
   vector<Data *> & getVarValues()
   {
      return varValues;
   }

   /** Gets the count of field values in this object
     * \return Returns the count of variable values
     */
   unsigned int numVars()
   {
      return varValues.size();
   }

   /** Gets a reference to the array of blank fields
     * \return Returns a reference to the array of blank fields
     */
   vector<string> & getBlankFields()
   {
      return blankFields;
   }

private:
   /** Flag to indicate different states of this object */
   static  unsigned int   contFlag;
   /** Name of this card object */
   string                 cardName;
   /** list of variable names in this card object */
   vector<string>         varNames;
   /** list of blank fields in this card object */
   vector<string>         blankFields;
   /** list of field values in this card object as generic data objects. The
     * generic data object stores different types of data like int, double, string
     */
   vector<Data *>         varValues;
};

/** Class to compare card names and find a card with a given name */
struct CompareString : std::binary_function<Card *, char *, bool>
{
public:
   /** Function overloaded for comparing the names of cards
     * \param card Pointer to the card object
     * \param card_name Name of the card to compare
     * \return Returns true if card names are same else false
     */
   bool operator()(Card *card, const char *card_name) const
   {
      string & kw = card->getName();

      if( !strcmp(kw.data(), card_name) )
        return true;
      else
        return false;
   }
};

/** Becareful before adding a member to this structure. This is referred as 3
  * ints in GetIDS class. If you change this, then getids.cpp has to be changed
  * accordingly.
  */
typedef struct _KWTypeIDStr
{
   /** group name of the keyword */
   unsigned int kwShort; 
   /** type of the keyword */
   unsigned int kwType; 
   /** ID of the entity or keyword */
   unsigned int id; 
}
  KWTypeIDStr, *KWTypeIDPtr;


/** Class to store the list of cards for different keywords of LSDYNA. The data 
  * from LSDYNA native file or from the database is stored in this object in
  * the form of cards. In turn the card stores the field names and their values.
  */
class ImportObject
{
public:
   /** Constructor to create this object with the given keyword name
     * \param keyword Name of the keyword
     */
   ImportObject(char *keyword)
   {
      keyWord = keyword;
      Card::setFlag();
      ioType = IMPORT_DYNA;
      cardCount = 0;
   }

   /** Constructor to create this object with the given keyword name
     * \param keyword Name of the keyword
     * \param kw_type Keyword type as an enum
     */
   ImportObject(char *keyword, unsigned int & kw_type)
   {
      keyWord = keyword;
      kwType = kw_type;
      Card::setFlag();
      ioType = IMPORT_DYNA;
      cardCount = 0;
   }

   /** Function to add a card object
     * \param card Card object pointer to add
     */
   void addCard(Card *card)
   {
      cardsList.push_back(card);
   }

   /** Function to get the keyword of this object
     * \return Returns a reference to the keyword of this object
     */
   string & getKeyword()
   {
      return keyWord;
   }

   /** Function to get the type of keyword of this object
     * \return Returns a reference to the keyword type of this object
     */
   unsigned int & getKeywordType()
   {
      return kwType;
   }

   /** Destructor to clear the memory of this object */
   virtual ~ImportObject()
   {
      vector<Card *>::iterator b = cardsList.begin(),e = cardsList.end();
      //Card *card;

      for(; b != e;b++)
      {
	 //card = *b;
         //if( card )
	    delete (*b);
      }

      cardsList.clear();
      refIDList.clear();
      IDList.clear();
      for(unsigned int i = 0; i < kwTypeIDList.size(); i++)
         delete kwTypeIDList[i];
      kwTypeIDList.clear();
   }

   /** Function to find a card with given name
     * \param card_name Name of the card to find in this object
     * \return Returns the Card object pointer if found or NULL if not found
     */
   Card *findCard(const char *card_name)
   {
      vector<Card *>::iterator b = cardsList.begin(),e = cardsList.end(),iter;

      iter = find_if(b, e, bind2nd( CompareString (), card_name));

      if( iter != e )
         return *iter;

      return NULL;
   }

   /** Function to save this object as a sequence of lines of data to a given 
     * file pointer
     * \param buf_ptr Pointer to the file
     */
   void toFile(BufFilePtr buf_ptr)
   {
      vector<Card *>::iterator b = cardsList.begin(),e = cardsList.end(),iter;
      Card *card;

      BufFileWrite(buf_ptr, (char *)keyWord.data(), keyWord.size());
      BufFileWrite(buf_ptr, (char *)"\n", 1);
      for(iter = b; iter < e; iter++)
      {
         card = *iter;
         card->toFile(buf_ptr);
      }
   }

   /** Function to save this object as a sequence of lines of data to a given 
     * string pointer
     * \param str Pointer to the string to which object's data is to be added
     */
   void tostring(string & str)
   {
      vector<Card *>::iterator b = cardsList.begin(),e = cardsList.end(),iter;
      Card *card;

      str += keyWord;
      str += "\n";
      for(iter = b; iter < e; iter++)
      {
         card = *iter;
         card->tostring(str);
      }
   }

   /** Function to get the reference to the list of cards in this object
     * \return Returns a reference to the list of cards
     */
   vector<Card *> & getCardList()
   {
      return cardsList;
   }

   /** Function to get the count of cards of this object. When saving the ids or
     * referred ids in a keyword, the data is not saved to this object as list of
     * cards. Then this function always returns 1.
     * \return Number of cards in this object
     */
   unsigned int length()
   {
      unsigned int num_cards = cardsList.size();
      if( num_cards )
         return num_cards;
      //this is added because when saving the ids and references to *.ids file 
      //or *_display.bin file, there won't be any cards saved to this object
      else if( Card::isFlagOn(SAVE_IDS|SAVE_REFERENCES) )
         return 1;
      else
         return 0;
   }

   /** Function to get the reference to the list of IDS if stored.
     * \return Pointer to the IDS list
     */
   vector<unsigned int> *getIDList()
   {
      return &IDList;
   }

   /** Function to get the reference to the list of keyword types and IDS
     * \return Pointer to the IDS and Keyword types
     */
   vector<KWTypeIDPtr> *getKWTypeIDList()
   {
      return &kwTypeIDList;
   }

   /** Function to set the group name for a list of keywords
     * \param kw_type Type of group
     */
   void setShortKW(unsigned int kw_type)
   {
      kwShort = kw_type;
   }

   /** Function to set the type of I/O. The I/O type can be read data from file 
     * or read data from database or export to dyna or export to database 
     * \param typ I/O type to set
     */
   void setIOType(int typ)
   {
      ioType = typ;
   }

   /** Function to check the status of the given flag
     * \return Returns 0 if not set otherwise an int
     */
   int isFlagOn(int typ)
   {
      return (ioType & typ);
   }

private:
   /** Type of keyword this object belongs to */
   unsigned int    kwType;
   /** Name of keyword this object belongs to */
   string          keyWord;
   /** List of cards added to this object */
   vector<Card *>  cardsList;

private:
   /** Type of group and flag to store the I/O type */
   unsigned int kwShort, ioType;
   /** Stores the list of IDS of referred entities in this keyword */
   vector<unsigned int> refIDList;
   /** Stores the list of IDS */
   vector<unsigned int> IDList;
   /** A list to store the KW type and ID of referred entities in a keyword */
   vector<KWTypeIDPtr> kwTypeIDList;
public:
   /** Count of number of cards in this object. This is used to assign the data
     * from database to the list of cards to show in Web interface 
     */
   unsigned int cardCount;
};

typedef map<string, ImportObject *>::iterator KWImpObjPtrIterator;
typedef std::pair<string, ImportObject *> KWImpObjPtrPair;

#endif

