#ifndef _DATA_LINE_H_
#define _DATA_LINE_H_

#define LINE_LEN 128
#define NUM_WORDS 8
#define WORD_LEN 10

#define LINE_LEN_MINUS_1 LINE_LEN-1
#define WORD_LEN_PLUS_2 WORD_LEN+2

#include <iostream>
#include "generic.h"

using namespace std;

typedef enum _KWFlag
{
   KW_READ_SUCCESS, /** Keyword's reading successful */
   KW_READ_FAIL, /** Keyword reading failed */
   KW_READ_START, /** Start of keyword data */
   KW_READ_END, /** End of keyword data */

   KW_UNKWOWN
}
  KWFlag;

/** Class to read a valid dataline from the given LSDYNA native file. Splits the
  * read line and stores their values in this object
  */
class DataLine
{
public:
   /** Constructor to create this object
     * \param fp File pointer to the LSDYNA native file
     */
   DataLine(FILE *fp)
   {
      filePtr = fp;
      flag = 0;
      m = 0;

      wordArray = new char *[NUM_WORDS];
      newWordArray = new char *[10];
      strBuff = new char[LINE_LEN];
      dblArray = new double[NUM_WORDS];

      for(m = 0; m < NUM_WORDS; m++)
         wordArray[m] = new char[WORD_LEN_PLUS_2];

      for(m = 0; m < 10; m++)
         newWordArray[m] = new char[24];
   }

   /** Function to read the next line of data from the given file
     * \param num_words Number of words to read and number of words successfully
     * read. This is both input and also output.
     * \return Returns the line of data read without splitting
     */
   inline char *readNextCard(unsigned int & num_words)
   {
      memset(strBuff, '\0', LINE_LEN);

      if( flag & (1 << KW_READ_END) )
      {
         num_words = 0;
         return strBuff;
      }

      flag = 0;
      while( fgets(strBuff, LINE_LEN, filePtr) )
      {
	 switch(strBuff[0])
	 {
	    case '$':
               memset(strBuff, '\0', LINE_LEN);
	       break;
	    case '*':
               num_words = 0;
	       flag |= (1 << KW_READ_END);
	       break;
	    default:
	       flag |= (1 << KW_READ_SUCCESS);
	       break;
	 }

	 if( (flag & (1 << KW_READ_SUCCESS)) || (flag & (1 << KW_READ_END)) )
	    break;
      }

      flag &= ~(1 << KW_READ_SUCCESS);

      if( strBuff[0] == '\0' )
         num_words = 0;

      return strBuff;
   }

   /** Function to read the next line of data from the given file
     * \return Returns the line of data read without splitting
     */
   inline char *readNextCard()
   {
      memset(strBuff, '\0', LINE_LEN);
      if( flag & (1 << KW_READ_END) )
      {
         return (char *)strBuff;
      }

      flag = 0;
      while( fgets(strBuff, LINE_LEN, filePtr) )
      {
	 switch(strBuff[0])
	 {
	    case '$':
               memset(strBuff, '\0', LINE_LEN);
	       break;
	    case '*':
	       flag |= (1 << KW_READ_END);
	       break;
	    default:
	       flag |= (1 << KW_READ_SUCCESS);
	       break;
	 }

	 if( (flag & (1 << KW_READ_SUCCESS)) || (flag & (1 << KW_READ_END)) )
	    break;
      }

      flag &= ~(1 << KW_READ_SUCCESS);

      return strBuff;
   }

   /** Function to read the next line of data from the given file, split it into
     * words and return them.
     * \param num_words Number of words to read and returns the number of words
     * successfully read
     * \return Returns the line of data as words or fields
     */
   inline char **readNextCardAsWords
                 (
                    unsigned int & num_words
                 )
   {
      memset(strBuff, '\0', LINE_LEN);

      if( flag & (1 << KW_READ_END) )
      {
         num_words = 0;
         return wordArray;
      }

      flag = 0;
      while( fgets(strBuff, LINE_LEN, filePtr) )
      {
	 switch(strBuff[0])
	 {
	    case '$':
               memset(strBuff, '\0', LINE_LEN);
	       break;
	    case '*':
               num_words = 0;
	       flag |= (1 << KW_READ_END);
	       break;
	    default:
               if( !strchr(strBuff, ',') )
               {
                  idx = 0;
		  for(m = 0; m < num_words; m++)
		  {
		    memset(wordArray[m], '\0', WORD_LEN_PLUS_2);
		    Generic::MemCpy(wordArray[m],&strBuff[WORD_LEN*m],WORD_LEN);
		    if( wordArray[m][0] != '\0' && wordArray[m][0] != '\n' )
		       idx++;
		  }
                  num_words = idx;
               }
               else
               {
                  m = 0;
                  char *str_char = strtok(strBuff, ",");
                  while( str_char )
                  {
		     memset(wordArray[m], '\0', WORD_LEN_PLUS_2);
                     Generic::MemCpy(wordArray[m], str_char, strlen(str_char));
                     m++;
                     str_char = strtok(NULL, ",");
                  }
                  num_words = m;
               }

	       flag |= (1 << KW_READ_SUCCESS);
	       break;
	 }

	 if( (flag & (1 << KW_READ_SUCCESS)) || (flag & (1 << KW_READ_END)) )
	    break;
      }

      flag &= ~(1 << KW_READ_SUCCESS);

      if( strBuff[0] == '\0' )
         num_words = 0;

#ifdef DEBUG
cout << "DATALINE = *****" << strBuff << "#####" << endl;
#endif

      return wordArray;
   }

   /** Function to read the next line of data from the given file, split it into
     * words, convert them into double data type and return them.
     * \param num_words Number of words to read and returns the number of words
     * successfully read
     * \return Returns the line of data as an array of doubles
     */
   inline double *readNextCardAsDoubles
                  (
                     unsigned int & num_words
                  )
   {
      memset(strBuff, '\0', LINE_LEN);
      if( flag & (1 << KW_READ_END) )
      {
         num_words = 0;
         return dblArray;
      }

      flag = 0;
      while( fgets(strBuff, LINE_LEN, filePtr) )
      {
	 switch(strBuff[0])
	 {
	    case '$':
               memset(strBuff, '\0', LINE_LEN);
	       break;
	    case '*':
               num_words = 0;
	       flag |= (1 << KW_READ_END);
	       break;
	    default:
               if( !strchr(strBuff, ',') )
               {
                  idx = 0;
		  for(m = 0; m < num_words; m++)
		  {
		     memset(wordArray[m], '\0', WORD_LEN_PLUS_2);
		     Generic::MemCpy(wordArray[m], &strBuff[WORD_LEN*m], 
							WORD_LEN);
#ifdef DEBUG
cout << "WORD = " << wordArray[m] << endl;
#endif
		     if( wordArray[m][0] != '\0' && wordArray[m][0] != '\n' )
		     {
			dblArray[idx] = atof(wordArray[m]);
                        idx++;
		     }
		  }
                  num_words = idx;
#ifdef DEBUG
cout << "NUM WORD=" << num_words << " NUMW=" << idx << endl;
#endif
               }
               else
               {
                  m = 0;
                  char *str_char = strtok(strBuff, ",");
                  while( str_char )
                  {
		     memset(wordArray[m], '\0', WORD_LEN_PLUS_2);
                     Generic::MemCpy(wordArray[m], str_char, strlen(str_char));
		     dblArray[m] = atof(wordArray[m]);
                     m++;
                     str_char = strtok(NULL, ",");
                  }
                  num_words = m;
               }
	       flag |= (1 << KW_READ_SUCCESS);
	       break;
	 }

	 if( (flag & (1 << KW_READ_SUCCESS)) || (flag & (1 << KW_READ_END)) )
	    break;
      }

      flag &= ~(1 << KW_READ_SUCCESS);

      if( strBuff[0] == '\0' )
         num_words = 0;

#ifdef DEBUG
cout << "DATALINE=*****" << strBuff << "#####" << endl;
#endif

      return dblArray;
   }

   /** Function to get a reference to the array of words stored in this object.
     * This object can be used to store the split words from the data line
     * \return Returns the pointer to the array of words.
     */
   inline char **getWordArray()
   {
      return wordArray;
   }

   /** Function to convert the given line of data into words and return them
     * as doubles.
     * \param str String to be converted to words and then doubles
     * \param num_words Count of words to be split
     * \return Returns the pointer to the array of doubles.
     */
   double *convertToDoubles(char *str, unsigned int & num_words)
   {
      for(m = 0; m < num_words; m++)
      {
	 memset(wordArray[m], '\0', WORD_LEN_PLUS_2);
	 Generic::MemCpy(wordArray[m], &str[WORD_LEN*m], WORD_LEN);
	 dblArray[m] = atof(wordArray[m]);
      }

      return dblArray;
   }

   /** Destructor to free the memory of this object */
   ~DataLine()
   {
      filePtr = NULL;
      for(m = 0; m < NUM_WORDS; m++)
         delete wordArray[m];

      for(m = 0; m < 10; m++)
         delete newWordArray[m];
      delete wordArray;
      delete newWordArray;
      delete strBuff;
      delete dblArray;
   }

   /** Sets a given flag 
     * \param _flag Flag to set
     */
   void setFlag(unsigned int _flag)
   {
      flag |= (1 << _flag);
   }

   /** Clears a given flag 
     * \param _flag Flag to clear
     */
   void clearFlag(unsigned int _flag)
   {
      flag &= ~(1 << _flag);
   }

   /** Checks a given flag for ON/OFF state
     * \param _flag Flag to check. If flag is not set, returns false 
     *  otherwise true
     */
   bool isFlagOn(unsigned int _flag)
   {
      if( flag & (1 << _flag) )
         return true;

      return false;
   }

   /** Clears all the flags by setting zero
     */
   void clearAll()
   {
      flag = 0;
   }

   /** Function to split the line of data based on widths of fields. Returns the
     * array of words
     * \param num_fields Count of fields to be split
     * \param field_widths Width for each field
     * \return Returns the array of words
     */
   char **splitIntoWords
          (
             unsigned int & num_fields,
             unsigned int *field_widths
          )
   {
      idx = 0;
      numFields = 0;
      for(m = 0; m < num_fields; m++)
      {
         if( field_widths[m] > 20 )
            break;
	 memset(newWordArray[m], '\0', 24);
         Generic::MemCpy(newWordArray[m], &strBuff[idx], field_widths[m]);
	 idx += field_widths[m];
         numFields++;
      }

      num_fields = numFields;

      return newWordArray;
   }

private:
   unsigned int      flag, m, idx, numFields;
   /** Holds the raw buffer read from the data file */
   char              *strBuff;
   //char     wordArray[NUM_WORDS][WORD_LEN+2];
   /** Holds the splitted words */
   char              **wordArray;
   /** Holds a copy of splitted words */
   char              **newWordArray;
   /** Stores an array of doubles */
   double            *dblArray;
   /** Pointer to the file */
   FILE              *filePtr;
};

#endif

