#ifndef _MANAGER_H_
#define _MANAGER_H_

#include <string>
#include "object.h"
#include "card.h"
#include "mantypes.h"

using namespace std;

typedef std::pair<string, unsigned int> StringUIntPair;
typedef std::multimap<string, unsigned int>::iterator StringUIntIterator;
typedef std::pair<int, int> IIPair;
typedef std::map<int, int> IIMap;
typedef std::map<int, int>::iterator IIIterator;

/** A base class to represent a Manager's functionality of managing the given 
  * data. Each group of keywords has a manager object to store the imported data.
  * The manager can be used to store, search and export the data to a file.
  * Whenever data is imported from database or from file, all keywords data
  * of a particular group is added to the respective manager. For example,
  * all keywords related to *AIRBAG are added to AirbagManager.
  */
class Manager : public Object
{
public:
   /** Constructor to create this object.
     * \param _idx Index of the ImportObject that is getting traversed. Default
     * is 0
     */
   Manager(unsigned int _idx = 0)
   {
      curIndex = _idx;
   }

   /** No functionality */
   virtual void tostring(char *)
   {
   }

   /** Resets the index to 0 for traversing the list */
   void resetIndex()
   {
      curIndex = 0;
   }

   /** Function to add a keyword object to its manager
     * \param kw_name Name of the keyword. Not used
     * \param imp_obj ImportObject to store data related to a keyword after
     * import from file or database
     */
   void add
        (
           const char *, 
           ImportObject *imp_obj
        )
   {
      impObjVector.push_back(imp_obj);
#ifdef DEBUG
cout << "ADDING OBJEC = " << imp_obj->getKeyword() << endl;
cout << "SIZE = " << impObjVector.size() << endl;
#endif
   }

   /** Virtual function for adding ID of object to manager */
   virtual void addID( char *, unsigned int) {} 
   /** Virtual function for finding ID of object in manager */
   virtual int findID(char *, unsigned int) {return 0;}

   /** Count of keyword objects in manager
     * \return Returns the count of keywords in this manager
     */
   unsigned int length()
   {
      return impObjVector.size();
   }

   /** Function to write all the objects in this to a given file pointer
     * \param buf_ptr File pointer to write the data in this object
     */
   void toFile(BufFilePtr buf_ptr)
   {
      vector<ImportObject *>::iterator b = impObjVector.begin(),
                                       e = impObjVector.end(), ptr;
      ImportObject *imp_obj;

      for(ptr = b; ptr != e; ptr++)
      {
         imp_obj = *ptr;
         imp_obj->toFile(buf_ptr);
      }
   }

   /** Function to append all the data in this object to a given string
     * \param str String pointer to append the data in this object
     */
   void tostring(string & str)
   {
      vector<ImportObject *>::iterator b = impObjVector.begin(),
                                       e = impObjVector.end(), ptr;
      ImportObject *imp_obj;

      for(ptr = b; ptr != e; ptr++)
      {
         imp_obj = *ptr;
         imp_obj->tostring(str);
      }
   }

   /** Function to get each object in the list as string
     * \param str String pointer to append the data in each object
     */
   void next_string(string & str)
   {
      if( !(impObjVector.size()) )
         return;

      ImportObject *imp_obj = impObjVector[curIndex];
      imp_obj->tostring(str);
      curIndex++;
      if( curIndex >= impObjVector.size() )
         curIndex = 0;
   }

   /** Destructor to free the memory in this object */
   virtual ~Manager()
   {
      vector<ImportObject *>::iterator b = impObjVector.begin(),
                                       e = impObjVector.end();
      for(; b != e; b++)
      {
//#ifdef DEBUG
      ImportObject *imp_obj = *b;
cout << "DELETING KW=" << imp_obj->getKeyword() << endl;
//#endif
	 delete (*b);
      }
      impObjVector.clear();
   }

   /** Function to get the reference to the list of ImportObject
     * \return Returns the reference to the list of objects
     */
   vector<ImportObject *> & getImportObjectList()
   {
      return impObjVector;
   }

public:
   /** Current index for traversing the list of objects */
   unsigned int           curIndex;
   /** Array of objects to store the data for keywords */
   vector<ImportObject *> impObjVector;
};

/*
void addIDToManager
     (
        char         **field_list,
        unsigned int id_index,
        unsigned int man_type,
        unsigned int & kw_type
     );

void findIDInManager
     (
        char         **id_list,
        unsigned int & num_indices,
        unsigned int *field_indices,
        unsigned int *man_types,
        unsigned int *kw_types
     );

bool validateMaterialForSectionType
     (
        int mat_type, 
        int sec_type
     );

void airbagFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void airbagVentHoleFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void aleFsiFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void airbagAdvancedAle
     (
        char         **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void aleRefGroupFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void aleUpSwitchFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void bemAcousticFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void bdrySphFlowFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void bdryPresMotionFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void bdrySpcFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void bdryTempFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void bdryThermalFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void constJointFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void contact2dAutoFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void contactFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void contactCouplingFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void contactEntityFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void contactGebodFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void contactGuidedFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void controlCurveFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void controlTemplateFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void controlUserFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void controlGeneralFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void databaseCSFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void databaseCSSetFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void databaseHistoryLocalFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void databaseHistoryIDFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void databaseHistoryFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void databaseFsiFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void defineAleBagFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void defineDTimesFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void defineDrawbeadFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void defineAdaptiveFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *
     );

void sbPretenFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void sbSlipringFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void elementMassFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void initialGasFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void initialTempFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void initialShellFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void initialVelGenFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void initialVelocityFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void initialPartFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void interfaceNodeFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void interfaceLinkFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void loadBeamFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void loadHeatGenFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void loadMovingPressureFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *
     );

void loadShellFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void loadNodeFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void sectionIntegrationFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void sensorElementFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void sensorControlFind
     (
        char **field_list,
        unsigned int & ,
        unsigned int id_index,
        ImportObject *
     );

void sensorForceFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void setNodeAddFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );

void setGenFind
     (
        char **field_list,
        unsigned int & num_fields,
        unsigned int id_index,
        ImportObject *imp_obj
     );
*/

#endif

