#ifndef _ALE_MANAGER_H_
#define _ALE_MANAGER_H_

#include "manager.h"
#include "aletype.h"

using namespace std;

/** Manager class to hold all the keywords starting with *ALE. All the keywords
  * are stored as list of ImportObject. This class holds its own instance here.
  */
class ALEManager : public Manager
{
private:
   /** Default constructor for creating this object */
   ALEManager() : Manager()
   {
   }

public:
   /** Static function to create this object. Object is created if already not 
     * there. only one instance exists at any point of time.
     */
   static ALEManager *create()
   {
      if( !managerPtr )
         managerPtr = new ALEManager();

      return managerPtr;
   }

   /** Destroys the stored instance of this object and initializes to NULL. */
   static void reset()
   {
      if( managerPtr )
         delete managerPtr;

      managerPtr = NULL;
   }

   ~ALEManager()
   {
      mmgIDTypeMap.clear();
      curveIDTypeMap.clear();
      grpIDTypeMap.clear();
      nodeIDTypeMap.clear();
      switchIDTypeMap.clear();
      fsiSwitchIDTypeMap.clear();

      managerPtr = NULL;
   }

   /** Adds a given object ID to this object with a given type
     * \param id ID of object to add
     * \param mtype Type of object to store id
     */
   void addID(char *id, unsigned int mtype)
   {
#ifdef DEBUG
cout << "ADDING ALE ID = " << id << endl;
#endif
      switch(mtype)
      {
	 case ALE_MULTI_MATERIAL_GROUP:
            {
               mmgIDTypeMap.insert(StringUIntPair(id, mtype));
            }
            break;
	 case ALE_FSI_SWITCH_MMG:
	 case ALE_FSI_SWITCH_MMG_ID:
	 case ALE_FSI_SWITCH_MMG_TITLE:
            {
               fsiSwitchIDTypeMap.insert(StringUIntPair(id, mtype));
            }
            break;
	 case ALE_REFERENCE_SYSTEM_CURVE:
            curveIDTypeMap.insert(StringUIntPair(id, mtype));
            break;
	 case ALE_REFERENCE_SYSTEM_GROUP:
            grpIDTypeMap.insert(StringUIntPair(id, mtype));
            break;
	 case ALE_REFERENCE_SYSTEM_NODE:
            nodeIDTypeMap.insert(StringUIntPair(id, mtype));
            break;
	 case ALE_REFERENCE_SYSTEM_SWITCH:
            switchIDTypeMap.insert(StringUIntPair(id, mtype));
            break;
      }
   }

   /** Function to find an object ID of given type
     * \param ale_map List of IDs and Types
     * \param id ID of object
     * \param mtype Not used
     * \param err_type Error message if any
     */
   int findIDBasedOnType
       (
          multimap<string, unsigned int> & ale_map,
          char *id, 
          unsigned int & ,
          string & err_type
       )
   {
      unsigned int num_ids = ale_map.count(id);

cout << "FINDID ALE = " << id << endl;
cout << "NUMIDS ALE = " << num_ids << endl;
#ifdef DEBUG
#endif
      switch(num_ids)
      {
         case 1:
            return 0;
         case 0:
	    cout << err_type << " ID <" << id << "> NOT FOUND" << endl;
	    return 1;
      }

      if( num_ids > 1 )
      {
	 StringUIntIterator low = ale_map.lower_bound(id), iter,
	                up = ale_map.upper_bound(id), e = ale_map.end();
         unsigned int cur_type, prev_type, num_clash = 1;
	  
         iter = low;
         prev_type = iter->second;
         iter++;
	 for(;iter != up; iter++)
	 {
            cur_type = iter->second;
	    if( cur_type == prev_type )
               num_clash++;
            prev_type = cur_type;
	 }

         if( num_clash > 1 )
         {
            cout << err_type << " ID is same for " << num_clash 
                 << " ales. ID = <" << id << ">" << endl;
         }
       
         return 0;
      }

      return 1;
   }

   /** Function to find an object ID of given type
     * \param id ID of object
     * \param mtype Type of object to find
     */
   int findID(char *id, unsigned int mtype)
   {
      string err_type;

      switch(mtype)
      {
         case FSI:
            err_type = "ALE_FSI_SWITCH_MMG";
            return findIDBasedOnType(fsiSwitchIDTypeMap, id, mtype, err_type);
         case MMG:
            err_type = "ALE_MULTI_MATERIAL_GROUP";
            return findIDBasedOnType(mmgIDTypeMap, id, mtype, err_type);
         case CURVE:
            err_type = "ALE_REFERENCE_SYSTEM_CURVE";
            return findIDBasedOnType(curveIDTypeMap, id, mtype, err_type);
         case GROUP:
            err_type = "ALE_REFERENCE_SYSTEM_GROUP";
            return findIDBasedOnType(grpIDTypeMap, id, mtype, err_type);
         case NOD:
            err_type = "ALE_REFERENCE_SYSTEM_NODE";
            return findIDBasedOnType(nodeIDTypeMap, id, mtype, err_type);
         case SWITCH:
            err_type = "ALE_REFERENCE_SYSTEM_SWITCH";
            return findIDBasedOnType(switchIDTypeMap, id, mtype, err_type);
         default:
            cout << "ALE of this type not exists" << endl;
            return 1;
      }
   }

private:
   /** An instance of this object */
   static ALEManager *managerPtr;
   /** A map holding the MMG ID and type */
   multimap<string, unsigned int> mmgIDTypeMap;
   /** A map holding the Switch ID and type */
   multimap<string, unsigned int> fsiSwitchIDTypeMap;
   /** A map holding the Curve ID and type */
   multimap<string, unsigned int> curveIDTypeMap;
   /** A map holding the group ID and type */
   multimap<string, unsigned int> grpIDTypeMap;
   /** A map holding the node ID and type */
   multimap<string, unsigned int> nodeIDTypeMap;
   /** A map holding the switch ID and type */
   multimap<string, unsigned int> switchIDTypeMap;
};

#endif

