#ifndef _Element_MANAGER_H_
#define _Element_MANAGER_H_

#include "manager.h"
#include "elementtype.h"

using namespace std;

/** Manager class to hold all the keywords starting with *ELEMENT. All the 
  * keywords are stored as list of ImportObject. This class holds its own 
  * instance here.
  */
class ElementManager : public Manager
{
private:
   /** Default constructor for creating this object */
   ElementManager() : Manager()
   {
   }

public:
   /** Static function to create this object. Object is created if already not 
     * there. only one instance exists at any point of time.
     */
   static ElementManager *create()
   {
      if( !managerPtr )
         managerPtr = new ElementManager();

      return managerPtr;
   }

   /** Destroys the stored instance of this object and initializes to NULL. */
   static void reset()
   {
      if( managerPtr )
         delete managerPtr;

      managerPtr = NULL;
   }

   ~ElementManager()
   {
      managerPtr = NULL;
   }

   /** Function to find an object ID of given type
     * \param id ID of object
     * \param mtype keyword type to find
     */
   int findID(char *id, unsigned int mtype)
   {
      multimap<string, unsigned int> *elm_map = NULL;

      switch(mtype)
      {
         case SHELL:
            elm_map = &shellIDTypeMap;
         case SOLID:
            elm_map = &solidIDTypeMap;
         case SBPRETEN:
            elm_map = &sbPreIDTypeMap;
         case SBSLIPRING:
            elm_map = &sbSlipIDTypeMap;
         case SBACCEL:
            elm_map = &sbAccelIDTypeMap;
         case SBELEMENT:
            elm_map = &sbIDTypeMap;
         case SBSENSOR:
            elm_map = &sbSensorIDTypeMap;
         case SBRETRACTOR:
            elm_map = &sbRetIDTypeMap;
         case BEAM:
            elm_map = &beamIDTypeMap;
         case DISCRETE:
            elm_map = &discreteIDTypeMap;
         case TSHELL:
            elm_map = &tshellIDTypeMap;
         case INERTIA:
            elm_map = &inertiaIDTypeMap;
         case MASS:
            elm_map = &massIDTypeMap;
         default:
            elm_map = &otherIDTypeMap;
      }

      unsigned int num_ids = elm_map->count(id);

#ifdef DEBUG
cout << "FINDID ELEMENT = " << id << endl;
cout << "NUMIDS ELEMENT = " << num_ids << endl;
#endif
      switch(num_ids)
      {
	 case 1:
	    return 0;
	 case 0:
	    cout << "ELEMENT ID <" << id << "> NOT FOUND" << endl;
	    return 1;
      }

      if( num_ids > 1 )
      {
	 StringUIntIterator low = elm_map->lower_bound(id), iter,
			up = elm_map->upper_bound(id), e = elm_map->end();
	 unsigned int cur_type, prev_type, num_clash = 1;
	  
	 iter = low;
	 prev_type = iter->second;
	 iter++;
	 for(;iter != up; iter++)
	 {
	    cur_type = iter->second;
	    if( cur_type == prev_type )
	       num_clash++;
	    prev_type = cur_type;
	 }

	 if( num_clash > 1 )
	 {
	    cout << "ID is same for " << num_clash << " elements. ID = <" 
		 << id << ">" << endl;
	 }
	 
	 return 0;
      }

      return 1;
   }

   /** Adds a given object ID to this object with a given type
     * \param id ID of object to add
     * \param kw_type Type of object to store id
     */
   void addID(char *id, unsigned int kw_type)
   {
      switch(kw_type)
      {
         case ELEMENT_SHELL:
         case ELEMENT_SHELL_BETA:
         case ELEMENT_SHELL_DOF:
         case ELEMENT_SHELL_MCD:
         case ELEMENT_SHELL_OFFSET:
         case ELEMENT_SHELL_SOURCE_SINK:
         case ELEMENT_SHELL_THICKNESS:
            {
               shellIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_BEAM:
         case ELEMENT_BEAM_OFFSET:
         case ELEMENT_BEAM_ORIENTATION:
         case ELEMENT_BEAM_PID:
         case ELEMENT_BEAM_SCALAR:
         case ELEMENT_BEAM_SCALR:
         case ELEMENT_BEAM_SECTION:
         case ELEMENT_BEAM_THICKNESS:
         case ELEMENT_BEAM_WARPAGE:
            {
               beamIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SEATBELT:
            {
               sbIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SEATBELT_ACCELEROMETER:
            {
               sbAccelIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SEATBELT_PRETENSIONER:
            {
               sbPreIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SEATBELT_RETRACTOR:
            {
               sbRetIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SEATBELT_SENSOR:
            {
               sbSensorIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SEATBELT_SLIPRING:
            {
               sbSlipIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_SOLID:
         case ELEMENT_SOLID_DOF:
         case ELEMENT_SOLID_ORTHO:
            {
               solidIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_TSHELL:
            {
               tshellIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_DISCRETE:
            {
               discreteIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_INERTIA:
         case ELEMENT_INERTIA_OFFSET:
            {
               inertiaIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_MASS:
         case ELEMENT_MASS_NODE_SET:
         case ELEMENT_MASS_PART:
         case ELEMENT_MASS_PART_SET:
            {
               massIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
         case ELEMENT_PLOTEL:
         case ELEMENT_SPH:
         case ELEMENT_TRIM:
         case ELEMENT_DIRECT_MATRIX_INPUT:
         default:
            {
               otherIDTypeMap.insert( StringUIntPair(id, kw_type) );
            }
            break;
      }
   }

private:
   static ElementManager *managerPtr;
   multimap<string, unsigned int> shellIDTypeMap;
   multimap<string, unsigned int> solidIDTypeMap;
   multimap<string, unsigned int> beamIDTypeMap;
   multimap<string, unsigned int> sbIDTypeMap;
   multimap<string, unsigned int> sbRetIDTypeMap;
   multimap<string, unsigned int> sbSensorIDTypeMap;
   multimap<string, unsigned int> sbAccelIDTypeMap;
   multimap<string, unsigned int> sbPreIDTypeMap;
   multimap<string, unsigned int> sbSlipIDTypeMap;
   multimap<string, unsigned int> discreteIDTypeMap;
   multimap<string, unsigned int> massIDTypeMap;
   multimap<string, unsigned int> inertiaIDTypeMap;
   multimap<string, unsigned int> tshellIDTypeMap;
   multimap<string, unsigned int> otherIDTypeMap;
};

#endif

