#ifndef _Set_MANAGER_H_
#define _Set_MANAGER_H_

#include "manager.h"
#include "settype.h"

using namespace std;

/** Manager class to hold all the keywords starting with *SET. All the keywords
  * are stored as list of ImportObject. This class holds its own instance here.
  */
class SetManager : public Manager
{
private:
   /** Default constructor for creating this object */
   SetManager() : Manager()
   {
   }

public:
   /** Static function to create this object. Object is created if already not 
     * there. only one instance exists at any point of time.
     */
   static SetManager *create()
   {
      if( !managerPtr )
         managerPtr = new SetManager();

      return managerPtr;
   }

   /** Function for deleting the current instance of this object and sets NULL
     * for the current instance */
   static void reset()
   {
      if( managerPtr )
         delete managerPtr;

      managerPtr = NULL;
   }

   ~SetManager()
   {
      managerPtr = NULL;
   }

   /** Adds a given object ID to this object with a given type
     * \param id ID of object to add
     * \param kw_type Type of object to store id
     */
   void addID(char *id, unsigned int kw_type)
   {
cout << "ADD SET ID " << id << endl;
      switch(kw_type)
      {
	 case SET_BEAM:
	 case SET_BEAM_ADD:
	 case SET_BEAM_GENERAL:
	 case SET_BEAM_GENERATE:
	    beamIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_DISCRETE:
         case SET_DISCRETE_ADD:
         case SET_DISCRETE_GENERAL:
         case SET_DISCRETE_GENERATE:
	    discreteIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_NODE:
         case SET_NODE_ADD:
         case SET_NODE_ADD_ADVANCED:
         case SET_NODE_COLUMN:
         case SET_NODE_GENERAL:
         case SET_NODE_LIST:
         case SET_NODE_LIST_GENERATE:
	    nodeIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
	 case SET_PART:
	 case SET_PART_ADD:
	 case SET_PART_COLUMN:
	 case SET_PART_LIST:
	 case SET_PART_LIST_GENERATE:
	    partIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_SEGMENT:
         case SET_SEGMENT_GENERAL:
	    segmentIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_SHELL:
         case SET_SHELL_ADD:
         case SET_SHELL_COLUMN:
         case SET_SHELL_GENERAL:
         case SET_SHELL_LIST:
         case SET_SHELL_LIST_GENERATE:
	    shellIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_SOLID:
         case SET_SOLID_ADD:
         case SET_SOLID_GENERAL:
         case SET_SOLID_GENERATE:
	    solidIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_TSHELL:
         case SET_TSHELL_GENERAL:
         case SET_TSHELL_GENERATE:
	    tshellIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
         case SET_MULTI_MATERIAL_GROUP_LIST:
	    mmgIDTypeMap.insert(StringUIntPair(id, kw_type));
            break;
      }
   }

   /** Function to find an object ID of given type
     * \param id ID of object
     * \param mtype keyword type to find
     */
   int findID(char *id, unsigned int mtype)
   {
      multimap<string, unsigned int>   *set_map = NULL;

      switch(mtype)
      {
         case BEAM:
            set_map = &beamIDTypeMap;
         case DISCRETE:
            set_map = &discreteIDTypeMap;
         case NOD:
            set_map = &nodeIDTypeMap;
         case PARTC:
            set_map = &partIDTypeMap;
         case SEGMENT:
            set_map = &segmentIDTypeMap;
         case SHELL:
            set_map = &shellIDTypeMap;
         case SOLID:
            set_map = &solidIDTypeMap;
         case TSHELL:
            set_map = &tshellIDTypeMap;
         case MMG:
            set_map = &mmgIDTypeMap;
         default:
            cout << "No set of this type exists" << endl;
            return 1;
      }

      unsigned int num_ids = set_map->count(id);

#ifdef DEBUG
cout << "FINDID SET = " << id << endl;
cout << "NUMIDS SET = " << num_ids << endl;
#endif
      switch(num_ids)
      {
         case 1:
            return 0;
         case 0:
	    cout << "SET ID <" << id << "> NOT FOUND" << endl;
	    return 1;
      }

      if( num_ids > 1 )
      {
	 StringUIntIterator low = set_map->lower_bound(id), iter,
	                up = set_map->upper_bound(id), e = set_map->end();
         unsigned int cur_type, prev_type, num_clash = 1;
	  
         iter = low;
         prev_type = iter->second;
         iter++;
	 for(;iter != up; iter++)
	 {
            cur_type = iter->second;
	    if( cur_type == prev_type )
               num_clash++;
            prev_type = cur_type;
	 }

         if( num_clash > 1 )
         {
            cout << "ID is same for " << num_clash << " sets. ID = <" 
                 << id << ">" << endl;
         }
       
         return 0;
      }

      return 1;
   }

private:
   static SetManager *managerPtr;
   multimap<string, unsigned int> beamIDTypeMap;
   multimap<string, unsigned int> discreteIDTypeMap;
   multimap<string, unsigned int> nodeIDTypeMap;
   multimap<string, unsigned int> partIDTypeMap;
   multimap<string, unsigned int> segmentIDTypeMap;
   multimap<string, unsigned int> shellIDTypeMap;
   multimap<string, unsigned int> solidIDTypeMap;
   multimap<string, unsigned int> tshellIDTypeMap;
   multimap<string, unsigned int> mmgIDTypeMap;
};

#endif

