#ifndef _CAE_CTX_H_
#define _CAE_CTX_H_

#include "caeenv.h"
#include "mantypes.h"
#include "card.h"

typedef insert_iterator< vector<unsigned int> > InsIter;

/** Class resembles a context of the CAE model. This is main class for creating
  * the model rendering and saving the model to an image
  */
class CAEContext
{
public:
   CAEContext() : caeEnv(NULL)
   {
   }

   /** Sets the transformations from the connection.xml. The connection.xml file
     * has all the details of transformations like rotation, translation and
     * scale. That information is set to the MTransform object.
     */
   void setTransform();
   /** Sets the view from the connection.xml. The connection.xml file
     * has the details of view like left, right, top, bottom, isometric and
     * front. That information is set to the MView object.
     */
   void setView();

   /** Set the parts to be switched on/off from the connection.xml */
   void setDisplay();
   /** Set the objects information other than parts from the connection.xml */
   void setObjectDisplay();

   /** This API is used to display parts separately as in *SET_PART */
   void setPartDisplay(const char *param_name);
   /** This API is used to display parts separately as in *SET_PART
     * \param on_off DISP_ON/DISP_OFF status of parts
     * \param obj_id_list List of part ids
     */
   void setPartDisplay(unsigned int on_off,vector<unsigned int> *obj_id_list);

   /** Main function for creating the CAEEnv object and sets the other 
     * transforms and views to the model. Both directory and model name are
     * used to locate the model information
     * \param dir Name of the directory
     * \param model Name of the model
     * \param width Width of model
     * \param height Height of model
     */
   void create
	(
	   const char *dir,
	   const char *model,
	   unsigned int width,
	   unsigned int height
	);

   /** Gets the object ids to be displayed from the connection.xml file. If
     * group name is element, then this returns the part ids for the elements.
     * \param kw_type_str Name of the keyword
     * \param id_list List of card indices from connection.xml
     * \param on_off Status for the objects
     * \param obj_id_list List of object ids to be displayed. This is filled here 
     * \param pid_list List of part ids for elements
     */
   void getIDs
	(
	   const char *kw_type_str,
	   string *id_list,
	   unsigned int *on_off,
	   vector<unsigned int> *obj_id_list,
	   vector<unsigned int> *pid_list = NULL
	);

   /** Function for displaying the *SET keywords
     * \param kw Name of the keyword
     * \param on_off DISP_ON/DISP_OFF status
     * \param obj_id_list List of entity ids to be displayed
     */
   void displaySets
	(
	   const char *kw,
	   unsigned int & on_off,
	   vector<unsigned int> & obj_id_list
	);

   ~CAEContext()
   {
      if( caeEnv )
         delete caeEnv;
      caeEnv = NULL;
   }

   /** For *SET with _GENERATE, this function expands the list of entities in
     * the set
     * \param id_list List of ranges of entity ids.
     * \param new_list The expanded list of entity ids are filled here
     */
   void expandList
        (
           vector<unsigned int> &id_list, 
           vector<unsigned int> *new_list
        );

   /** Displays node ids with the given status
     * \param on_off DISP_ON/DISP_OFF status of nodes
     * \param obj_id_list List of node ids to be displayed
     */
   void displayNodes
	(
	   unsigned int on_off,
	   vector<unsigned int> & obj_id_list
	);

   /** Function for displaying entities other than parts, nodes, elements and sets
     * \param on_off DISP_ON/DISP_OFF status for the objects
     * \param kw_ptr Pointer holding the referred entities ids and their types
     */
   void displayGenericObjects
	(
	   unsigned int on_off,
	   KWTypeIDStr *kw_ptr
	);

   void renderModel()
   {
       if(caeEnv)
       {
           caeEnv->renderModel();
           glFinish();
       }
   }

private:
   const char *dirName, *modelName;
   CAEEnv *caeEnv;
};

#endif

