#ifndef _READ_ELEMENTS_H_
#define _READ_ELEMENTS_H_

#include <set>
#include <map>
#include "readids.h"
#include "readnodes.h"
#include "partstr.h"

#define UNSIGNED_INT_MAX_FLAG (((unsigned int)1)<<31)

using namespace std;

/** Struct representing *NODE keyword */
typedef struct _NodeStr
{
   /** Index of node */
   unsigned int index;
   /** x, y and z coordinates */
   float coords[3];
}
  NodeStr, *NodePtr;

/** Class for reading the elements information from the binary files. */
class ReadElements
{
public:
   /** Constructor for creating this object
     * \param pid_map Map of part id and its index is filled in this pointer
     */
   ReadElements(UIUIMap *pid_map, vector<ReadIDS *> *id_vect, vector<ReadNodes *> *node_vect);

   ~ReadElements();

   /** Constructor for creating this object
     * \param dir Name of the directory from which the files are loaded
     * \param model Name of the model to load
     * \param file_idx Index of file when include files are there
     * \param pid_map Pointer to map for part ids and indices
     */
   ReadElements(const char *dir, const char *model,
                                             int file_idx, UIUIMap *pid_map);

   /** Values are to this class
     * \param dir Name of the directory
     * \param model Name of the model to load
     * \param file_idx Index of file when include files are there
     * \param read_ids Pointer to object of reading IDs
     * \param read_nodes Pointer to object of reading nodes and their values
     */
   void setValues(const char *dir, const char *model, int file_idx,
                                ReadIDS *read_ids, ReadNodes *read_nodes);

   /** Generates a random color and gets the color generated
     * \param rgb Array to hold the color values
     */
   static inline void randomColor(float rgb[3]);

   /** Gets the elements information along with the parts
     * \param type Type of element.shell, solid, beam, disc, sb, tshell, sbaccel
     * \param num_parts Number of parts for the model
     * \param parts Pointer to list of parts information
     */
   int get(const char *type, unsigned int *num_parts, PartInfoStr **parts);

   /** Get the elements and parts information for given list of part ids
     * \param pid_list List of part ids for which the information to be filled
     * \param num_parts Number of parts for which the information is filled
     * \param parts Pointer to the list of parts info
     * \return Returns 0 if failed else greater than 0 
     */
   int getPartsByID
       (
	  vector<unsigned int> *pid_list,
	  unsigned int *num_parts,
	  PartInfoStr  **parts
       );

   /** Get the part type based on element type
     * \param type Type of element likw shell, solid, disc, beam, seatbelt,
     * tshell, mass, inertia
     * \return Returns the type of part
     */
   unsigned int getPartType(const char *type);

   /** This function loads all the elements into memory from the binary files
     * This function inturn calls the get function for each type of element
     * \param num_parts Number of parts loaded
     * \param parts Pointer to hold the loaded information of parts
     * \return Returns total number of elements loaded
     */
   int getAll(unsigned int *num_parts, PartInfoStr **parts);

   /** For each part unique indices of nodes are stored. All the element nodes
     * are searched in this list. If node is found then index of that node is
     * saved with part. If not found, then that element is removed from the
     * list of part elements
     * \param no_node_flag While loading elements, this flag is set. If all nodes
     * are found for the part then, this flag is true else false
     * \param part_info Information of part for which the elements are to be
     * checked for node details and nodes presence
     */
   void fillNodeDetails(unsigned int no_node_flag, PartInfoPtr part_info);

   /** This function creates a temporary segment part for all elements of the 
     * segment based on the node list given. For segment, the solid element
     * faces can also exist. So a separate part is created and displayed. But
     * this has to be improved to display the faces of solid elements and quads
     * of certain parts
     * \param nid_list List of nodes for the segment. First four nodes becomes
     * an element and next four another. All nodes of each element is searched
     * in parts and element is created for these nodes and added to this part.
     */
   PartInfoPtr createPartForSegment
       (
	  vector<unsigned int> & nid_list
       );

private:
   /** Name of directory and model name */
   const char                         *dirName, *modelName;
   /** Index of file */
   int                                fileIndex;
   /** Pointer to the obejct which reads entity ids and their card indices */
   ReadIDS                            *readIDS;
   /** Pointer to the obejct which reads node details */
   ReadNodes                          *readNodes;
   /** Uniq nodes of each part is stored here */
   map<unsigned int, unsigned int>    uniqIndices;
   /** Part id and its index map */
   UIUIMap                            *pidMap;
   vector<ReadIDS *>                  *m_ReadIDVect;
   vector<ReadNodes *>                *m_ReadNodeVect;
};

#endif

