#ifndef _CAE_ENV_H_
#define _CAE_ENV_H_

#include "partstr.h"
#include "mview.h"
#include "mtrans.h"
#include "gentime.h"
#include "trackball.h"
#include "boundbox.h"
#include "maptypes.h"
#include "mantypes.h"
#include <string>
#include <iostream>
#include <vector>
#include <set>
#include <GL/glext.h>
#include "readnodes.h"
#include "readids.h"

using namespace std;

/** Represents a quad element */
typedef struct _Quad
{
   /** four node ids of a Quad element */
   unsigned int nodes[4];
}
  Quad, *QuadPtr;

/** Type of display */
typedef enum _DisplayType
{
   WIRE_FRAME,
   SHADED,
   SHADED_WIRE_FRAME,
   SHADED_TRANSPARENT,
   SHADED_WIRE_FRAME_TRANSPARENT,
   HIDDEN_SURFACE_REMOVAL,

   DISPLAY_TYPE_MAX
}
  DisplayType;

const unsigned int NUM_COORDS = 3;

/** Very important class to do the model rendering. The actual rendering is
  * done in this object. Resembles a CAE environment object to show the nodes
  * and element connectivity as a model. It displays shell, solid and beam
  * elements in different forms of display.
  */
class CAEEnv
{
public:
   /** Constructor for creating this object
     * \param dir_name Name of the directory
     * \param model_name Name of model has to be loaded
     * \param shade Type of shading to display. Default is GL_SMOOTH
     */ 
   CAEEnv(const char *dir_name, const char *model_name, int shade = GL_SMOOTH);
   /** Gets the Model View pointer hold by this object
     * \return Model view pointer
     */
   MView *getMViewPtr();
   /** Gets the Model transform pointer hold by this object
     * \return Model view pointer
     */
   MTransform *getMTransformPtr();
   /** Gets the number of parts
     * \return count of parts
     */
   int getNumParts();
   /** Sets the type of pick whether node, element or part
     * \param _flag Type of pick
     */
   void setPickType(unsigned int _flag);
   /** Function to load the model data from binary files of node and elements
     * \param pid_list The list of parts to be loaded. Based on requirement
     * the list can be specified. One can load all the parts or some parts.
     * If NULL, then all parts are loaded
     * \return Returns 0 on success else greater than 0
     */
   int getAll(vector<unsigned int> *pid_list=NULL);

   /** Loads all the binary files and initializes the GL
     * \param font_base Font base ID for drawing the box and its dimensions
     * \param thread_status Status of thread. The loading of binary files
     * is given to a thread. Based on the status of the thread, the files
     * are loaded. If thread status is true, then files are not loaded again. If
     * not then files are loaded without a thread.
     * \param pid_list List of parts to be loaded. If NULL, all parts are loaded
     */
   void loadAndInitGL(unsigned int font_base,bool thread_status = false,
                         vector<unsigned int> *pid_list = NULL);

   /** Function called initially and after resize of model.
     * \param w Width of window
     * \param h Height of window
     */
   void resizeModel(int w, int h);

   /** Function to call for rendering of the model */
   void renderModel();

   /** Switches on the lights. There are two lights for the model rendering */
   void switchOnLights();

   /** Sets type of shade to the model
     * \param shade Type of shade GL_FLAT or GL_SMOOTH
     */
   void setShadeModel(unsigned int shade);

   /** The node and element normals are computed based on the need. For smooth
     * shading the node normals are computed. But flat shading the element 
     * normals are computed
     * \param flag Flag to indicate whether element normals are computed or not.
     * To display selected nodes and to display segments, this flag can be used 
     */
   void setComputeFlag(unsigned int flag);
   void clearComputeFlag(unsigned int flag);
   int checkComputeFlag(unsigned int flag);

   /** The parts are processed to switch on/off. 
     * \param on_off DISP_ON or DISP_OFF state
     */
   void processParts(unsigned int on_off);

   /** Function for switching on all the parts */
   void displayAllParts();

   /** Function for rendering the parts by element. When some elements are
     * picked or selected, then this function is used.
     */
   void displayPartsByElement();

   /** This is a regular function for rendering the parts. This works based
     * on list of elements to render. So this is faster than displayPartsByElement
     * function. The parts can be displayed based on file index when include
     * files are there in the model
     * \param file_index Index of file for which the parts are to be displayed.
     * Default is -1. Indicates all the parts from all files are to be displayed. 
     */
   void displayParts(int file_index = -1);

   /** Faster function for rendering the part having only shell elements
     * \param part_info The information of part to render
     * \param color Color of part to render
     */
   void renderShellPart(PartInfoPtr part_info, float *color);

   /** Function for rendering the part for pick of shell elements
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderShellsForPick
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part for pick of discrete elements
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderDiscretesForPick
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Faster function for rendering the part having only discrete elements.
     * \param part_info The information of part to render
     * \param color Color of part to render
     */
   void renderDiscretePart
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part for pick of discrete elements. Individual
     * elements are rendered
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderDiscretes
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part for pick of beam elements
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderBeamsForPick
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Faster function for rendering the part having beam elements.
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderBeamPart
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part having beam elements to process for pick.
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderBeams
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part having shell elements to process for pick.
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderShells
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Faster function for rendering the part having solid elements.
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderSolidPart
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part having solid elements to process for pick
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderSolidsForPick
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering the part having solid elements to process for pick
     * \param part_info The information of part to render
     * \param color Part color for rendering
     */
   void renderSolids
               (
                  PartInfoPtr part_info,
                  float *color
               );

   /** Function for rendering Tria element by its node normal. This is for SMOOTH
     * rendering
     * \param indices Indices of Tria element to render
     * \param node_values Node coordinates for tria vertex nodes
     * \param node_normals Node normals for tria
     */
   void renderTriaByNodeNormal
        (
           unsigned int *indices, 
           float *node_values, 
           float *node_normals
        );

   /** Function for rendering quad element by its node normal. This is for SMOOTH
     * rendering
     * \param indices Indices of quad element to render
     * \param node_values Node coordinates for quad vertex nodes
     * \param node_normals Node normals for quad
     */
   void renderQuadByNodeNormal
        (
           unsigned int *indices, 
           float *node_values, 
           float *node_normals
        );

   /** Function for rendering Tria element by its normal. This is for FLAT
     * rendering
     * \param indices Indices of Tria element to render
     * \param node_values Node coordinates for tria vertex nodes
     * \param normal_ptr Node normals for tria
     */
   void renderTriaByNormal
        (
           unsigned int *indices, 
           float *node_values, 
           float *normal_ptr
        );

   /** Function for rendering quad element by its normal. This is for FLAT
     * rendering
     * \param indices Indices of quad element to render
     * \param node_values Node coordinates for quad vertex nodes
     * \param normal_ptr Node normals for quad
     */
   void renderQuadByNormal
        (
           unsigned int *indices, 
           float *node_values, 
           float *normal_ptr
        );

   /** Function for rendering tetra element by its normal. This is for FLAT
     * rendering
     * \param indices Indices of tetra element to render
     * \param norm_indices Indices of tetra element normals
     * \param elm_normals Element normal for tetra
     * \param node_values Node coordinates for tetra vertex nodes
     */
   void renderTetraByNormal
        (
           unsigned int *indices,
           unsigned int *norm_indices, 
           float *elm_normals, 
           float *node_values
        );

   /** Function for rendering penta element by its normal. This is for FLAT
     * rendering
     * \param indices Indices of penta element to render
     * \param norm_indices Indices of penta element normals
     * \param elm_normals Element normal for penta
     * \param node_values Node coordinates for penta vertex nodes
     */
   void renderPentaByNormal
        (
           unsigned int *indices,
           unsigned int *norm_indices, 
           float *elm_normals, 
           float *node_values
        );

   /** Function for rendering hexa element by its normal. This is for FLAT
     * rendering
     * \param indices Indices of hexa element to render
     * \param norm_indices Indices of hexa element normals
     * \param elm_normals Element normals for hexa
     * \param node_values Node coordinates for hexa vertex nodes
     */
   void renderHexaByNormal
        (
           unsigned int *indices,
           unsigned int *norm_indices, 
           float *elm_normals, 
           float *node_values
        );

   /** Function for rendering tetra element by nodes normal. This is for SMOOTH
     * rendering
     * \param indices Indices of tetra element to render
     * \param node_values Nodes coordinates for tetra
     * \param node_normals Nodes normal for tetra
     */
   void renderTetraByNodeNormal
        (
           unsigned int *indices,
           float *node_values,
           float *node_normals
        );

   /** Function for rendering penta element by nodes normal. This is for SMOOTH
     * rendering
     * \param indices Indices of penta element to render
     * \param node_values Nodes coordinates for penta
     * \param node_normals Nodes normal for penta
     */
   void renderPentaByNodeNormal
        (
           unsigned int *indices,
           float *node_values,
           float *node_normals
        );

   /** Function for rendering hexa element by nodes normal. This is for SMOOTH
     * rendering
     * \param indices Indices of hexa element to render
     * \param node_values Nodes coordinates for hexa
     * \param node_normals Nodes normal for hexa
     */
   void renderHexaByNodeNormal
        (
           unsigned int *indices,
           float *node_values,
           float *node_normals
        );

   ~CAEEnv();

   /** Gets the pointer to the array having the model extents
     * \return Pointer the array
     */
   float *getExtents();

   /** Gets pointer to the name of model
     * \return Returns the pointer to the name of model
     */
   const char *getModelName();

   /** Gets the instance of this class pointer
     * \return Returns the pointer to this class instance
     */
   static CAEEnv *getPtr();
   /** Sets the instance of this class pointer
     * \return the pointer to this class instance
     */
   static void setPtr(CAEEnv *ptr);

   /** Computes the normals to nodes and elements */
   void computeNormals();
   /** Computes the shell normals */
   void computeShellNormals(PartInfoPtr part_info);
   /** Computes the Solid normals */
   void computeSolidNormals(PartInfoPtr part_info);

   /** Computes both Shell and Solid element normals
     * \param nodes Indices of nodes of elements
     * \param node_values Coordinates of nodes of elements
     * \param node_normals Computed list of node normals
     * \param elm_normals Computed list of element normals
     */
   void computeElementNormal
        (
           unsigned int *nodes, 
           float *node_values, 
           float *node_normals, 
           float *elm_normals
        );

   /** Select a part based on given part index
     * \param part_num Index of part to select
     */
   void selectPart(unsigned int part_num);

   /** Select parts by their ids
     * \param id_list List of part ids to select
     */
   void selectPartsByID(vector<unsigned int> *id_list);

   /** Highlight a part based on given part index
     * \param part_num Index of part to highlight
     */
   void highlightPart(unsigned int part_num);

   /** Sets on/off status to a list of parts in one call
     * \param num_parts Count of parts
     * \param parts_list List of parts to set status
     * \param on_off DISP_ON/DISP_OFF status
     */
   void setDisplay(unsigned int num_parts, unsigned int *parts_list, 
               unsigned int on_off);
   /** Sets the display of the given part index
     * \param part_num Index of part to set the display status
     * \param on_off DISP_ON/DISP_OFF status
     */
   void setDisplay(unsigned int part_num, unsigned int on_off);

   /** Clear the select display for the parts */
   void clearSelectedParts();

   /** Function to set the display of box around parts to DISP_ON/DISP_OFF */
   void setBoxDisplay(int box_on_off);

   /** Save the rendering data like shading, switched on/off parts, view etc to
     * a binary file for restoring. Using this funtion, save of a state of model
     * can be done to restore it at any point of time. This function is used
     * for stand alone application saving.
     * \param file_name Name of the file to save the data
     */
   int save(const char *file_name);
   /** Restore the rendering data like shading, switched on/off parts, view etc
     * from binary file previously saved. The state of model at any point of time
     * can be restored using this function. This function is used for stand alone 
     * application restore.
     * \param file_ptr Pointer to the file to restore
     */
   void restore(FILE *file_ptr);

   /** Select an element for display when element is picked
     * \param part_idx Index of part to which this element belongs
     * \param elm_idx Index of element to display. Each element has an index in
     * a part while loading the binary data from file. The index of element
     * can be found element id. There is a map between the element id and index.
     */
   void selectElement(unsigned int *part_idx, unsigned int *elm_idx);

   /** Selects a list of elements based on their ids and their respective part
     * ids.
     * \param eid_list List of element IDs to select
     * \param pid_list List of part IDs holding the above elements
     */
   void selectElements(vector<unsigned int> & eid_list,
                                         vector<unsigned int> & pid_list);

   /** Selects a list of elements based on their indices and their respective part
     * ids.
     * \param eis_list List of element indices to select
     * \param pis_list List of part IDs holding the above elements
     */
   void selectElementsByIndex(vector<unsigned int> & eis_list,
                                         vector<unsigned int> & pis_list);
   /** Clears selected state for already selected elements */
   void clearSelectedElements();

   /** Selects a node based on its index and index of the part it resides
     * \param part_idx Index of the part to which the node belongs to
     * \param node_idx Index of the node to select
     */
   void selectNode(unsigned int *part_idx, unsigned int *node_idx);

   /** Select a list of nodes based on their ids
     * \param nid_list List of node ids to select. When this list is big, this
     * function is going to take lot of time. Each node id is searched in each
     * part and then selected if found. With this method only the part nodes can 
     * be selected but not the non-part nodes and also this is taking more than 
     * a minute for 1.25 millon nodes selection
     */
   void selectNodes(vector<unsigned int> & nid_list);

   /** Clears the select state for the already selected nodes */
   void clearSelectedNodes();

   /** Prints the information of node
     * \param idx Index of node to print info
     * \param n Coordinate values
     */
   void printNode(unsigned int idx, float *n);

   /** Sets blending or no blending for transparent display
     * \param flag Flag for blending or noblending
     */
   void setBlend(int flag = 1);

   /** Sets hidden line removal or no hidden line removal 
     * \param flag Flag for hidden line removal
     */
   void setHideMode(int flag = 1);

   /** Sets display to wireframe or filled
     * \param mode Flag for displaying wireframe or filled
     */
   void setPolygonMode(int mode = GL_FILL);

   /** Normalizes the node values
     * \param part_info Part information to get node coordinates
     */
   void normalize(PartInfoPtr part_info);

   /** This function computes the extents of the displayed objects and updates
     * the view volume
     */
   void modifyExtents();
 
   /** Function to change the position of lights based on the new extents of 
     * the displayed objects
     */
   void changeLightPositions();

   /** Allocates memory for selected elements to display. As the color has to be
     * chaged for the selected elements, the list is stored separately. The list
     * is created as and when required only
     */
   void createElementSelectedList();

   /** Allocates memory for selected nodes to display. As the color has to be
     * chaged for the selected nodes, the list is stored separately. The list
     * is created as and when required
     */
   void createNodeSelectedList();

   /** Get the pointer to the list of parts info
     * \return Returns the pointer to the list of parts info
     */
   PartInfoPtr getPartsList();

   /** A map is created between the part ids and indices. This function gets
     * map pointer
     * \return Returns the pointer to the map
     */
   UIUIMap *getPartsMapPtr();

   /** This function can used for selecting the non-structured nodes also. When
     * all the nodes of a model is to be selected for creating a set or any
     * object, then this function can be used. This is a faster function.
     */
   void selectNodeList(vector<unsigned int> & nid_list);

   /** When displaying the segments, all the segment elements are added to a
     * temporary part and that part is set to this object for display.
     * \param part_info Part information
     */
   void setSegmentPart(PartInfoPtr part_info);

   /** Switch On/Off the parts based on the given flag. Parts are processed
     * which to swicth on or which parts to switch off.
     * \param flag For to indicate DISP_ON/DISP_OFF
     * \param id_list List of part ids for processing
     */
   void processPartsByID(unsigned int flag, vector<unsigned int> *id_list);

private:
   char         charBuff[128];
   /** Selected node list for non-structured nodes */
   char         *selectedNodes;
   /** Total number of parts in the model */
   unsigned int numParts;
   /** Type of shape to draw GL_QUAD or GL_TRIANGLE. Tried for making the 
     * rendering fast. This is not used now */
   unsigned int primitiveType; //quad or triangle
   unsigned int i, compFlag;
   unsigned int pickType;
   /** Total number of nodes in the model */
   unsigned int numNodes;
   /** OpenGL parameters for shading, blending and Fill */
   int          shadeModel, blendFlag, polygonMode[2];
   int          viewPort[4];
   int          numTrans;
   /** Extents of model xmin, xmax, ymin, ymax, zmin, zmax */
   float        modelExtents[6];
   /** copy of model extents */
   float        orgModelExtents[6];
   /** Blue for select and green for highlight */
   float        colorBlue[4], colorGreen[4];
   /** Node coordinates for entire model. */
   float        *nodeValues;
   FILE         *filePtr;
   /** Parts information about nodes and element connectivity */
   PartInfoStr  *partsInfoList;
   /** A Temporary part for segment display */
   PartInfoPtr  segmentPart;
   /** Object pointer for different views of model */
   MView        *mView;
   /** Object pointer for different transformations of model */
   MTransform   *mTrans;
   /** An instance of this class */
   static CAEEnv *cae_Env;
   /** Pointer to hold the list of boxes created around the parts */
   BoundBox     **boxList;
   /** Pointer to ids and card indices mapping object */
   ReadIDS      *readIDS;
   /** Pointer to nodes reading object */
   ReadNodes    *readNodes;
   /** A map between the part id and its index */
   UIUIMap      pidMap;
   /** Directory from which the model is loaded */
   string       dirName;
   /** Name of model loaded */
   string       modelName;
   /** List of include files with the main file */
   vector<string> fileList;
};

#endif

