#ifndef _M_TRANS_H_
#define _M_TRANS_H_

#include <GL/gl.h>
#include <GL/glu.h>
#include "trackball.h"
#include "rubberband.h"
#include "unit_axis.h"
#include <iostream>
#include <string.h>

#define MAX3(a,b,c) ((a > b) ? ((a > c) ? a : c) : ((b > c) ? b : c))
#define MIN3(a,b,c) ((a < b) ? ((a < c) ? a : c) : ((b < c) ? b : c))

using namespace std;

const double Z_MAX_COORD=9999.0f;
const double Z_MIN_COORD=-9999.0f;
const float SCALE_FACTOR=0.2f;
const float MIN_SCALE_FACTOR=0.1f;

#define DISP_NORMAL           (1 << 0)
#define DISP_ON               (1 << 1)
#define DISP_OFF              (1 << 2)
#define DISP_SELECT           (1 << 3)
#define DISP_PARTIAL          (1 << 4)
#define DISP_HIGHLIGHT        (1 << 5)
#define DISP_BOX              (1 << 6)
#define DISP_SEGMENT          (1 << 7)
#define DISP_CLEAR            (1 << 8)

#define PART_PICK         (1 << 10)
#define ELEMENT_PICK      (1 << 11)
#define NODE_PICK         (1 << 12)
#define SELECT_NODES_ONLY (1 << 13)

#define COMPUTE_ELEMENT_NORMALS   (1 << 20)
#define CREATE_SELECT_ELEMENTS    (1 << 21)
#define CREATE_SELECT_NODES       (1 << 22)
#define COMPUTE_NODE_NORMALS      (1 << 23)

#define FILL_NODE_DETAILS     (1 << 25)
#define FILL_ELEMENT_DETAILS  (1 << 26)

/** Transformation types */
typedef enum _TransType
{
   NO_TRANSFORM,
   RESET_TRANSFORM,
   ROTATE_X_CCW,
   ROTATE_X_CW,
   ROTATE_Y_CCW,
   ROTATE_Y_CW,
   ROTATE_Z_CCW,
   ROTATE_Z_CW,

   TRANS_LEFT,
   TRANS_RIGHT,
   TRANS_UP,
   TRANS_DOWN,

   SCALE_INCR,
   SCALE_DECR,
   GLOBAL_TRANSLATION,
   GLOBAL_SCALE,
   GLOBAL_ROTATION,
   START_RUBBER_BAND,
   START_POLYLINE,
   END_RUBBER_BAND,
   RECT_SCALE,
   RECT_PICK,
   SELECT_PICK,
   POLYLINE_PICK,

   TRANS_MAX
}
  TransType;

typedef enum _MouseTypes
{
   BUTTON1,
   BUTTON2,
   BUTTON3,

   BUTTON_PRESS,
   BUTTON_RELEASE,
   MOTION_NOTIFY,
}
  MouseTypes;

/** Class for representing the model transformations like rotation, translation,
  * scale along the x, y and z axes
  */
class MTransform
{
public:
   /** Constructor for creating this object
     * \param exts Pointer to the model extents array
     */
   MTransform(float *exts) : mTransType(NO_TRANSFORM), onlyOnce(DISP_ON),
         modelExtents(exts)
   {
      resetAll();
      axisLength[0] = ( ((exts[0] >= 0) ? exts[0] : exts[0] * -1) + 
                    ((exts[1] >= 0) ? exts[1] : exts[1] * -1) )/15.0f;
      axisLength[1] = ( ((exts[2] >= 0) ? exts[2] : exts[2] * -1) + 
                    ((exts[3] >= 0) ? exts[3] : exts[3] * -1) )/10.0f;
      axisLength[2] = ( ((exts[4] >= 0) ? exts[4] : exts[4] * -1) + 
                    ((exts[5] >= 0) ? exts[5] : exts[5] * -1) )/10.0f;
cout << "AXIS LEN=" << axisLength[0] << "," << axisLength[1] << "," << axisLength[2] << endl;
   }

   /** Initializes the arrays of values to zero or 1 */
   void resetAll();

   /** Function for changing the transformation type *
     * \param trans_type Type of transformation
     */
   void changeTransform(int trans_type)
   {
      mTransType = trans_type;
   }
   
   /** Updating the transformation values
     * \param trans_type Transformation type
     * \param values Array of values to be modified
     */
   void updateValues(int trans_type, float values[4]);

   /** Setting array of values for tranformation. First 3 values global rotation
     * next 3 values global transformation and next 3 values for scaling
     * \param values Array of values to be set
     */ 
   void setValues(float values[9]);

   ~MTransform()
   {
   }

   /** Updates the global arrays from the given values */
   void update();

   /** Sets the rotation values
     * \param values first 3 values are the direction of rotation vector and
     * fourth value is the angle of rotation
     */
   void setRotation(float values[4]);

   /** Sets the translation values
     * \param values 3 values are the translation along x, y and z directions
     */
   void setTranslation(float values[3]);

   /** Thsi is the first function to call to really render the model. After 
     * calling this function, the GL APIS are called to really draw the model.
     */
   inline void processStart()
   {
      update();
      glGetDoublev(GL_MODELVIEW_MATRIX, mvMatrixLocal);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
         mvMatrixLocal[12] += globalTrans[0];
	 mvMatrixLocal[13] += globalTrans[1];
	 mvMatrixLocal[14] += globalTrans[2];
	 glLoadMatrixd(mvMatrixLocal);

fprintf(stderr, "GTRAN=%lf %lf %lf\n", globalTrans[0], globalTrans[1], globalTrans[2]);
fprintf(stderr, "GTRAN=%lf %lf %lf\n", globalRot[0], globalRot[1], globalRot[2]);
	 glPushMatrix();
	    glTranslatef(centerOfWindow[0], centerOfWindow[1], 
                                            centerOfWindow[2]);
	    glRotatef(globalRot[0], 1.0, 0.0, 0.0);
	    glRotatef(globalRot[1], 0.0, 1.0, 0.0);
	    glRotatef(globalRot[2], 0.0, 0.0, 1.0);
            build_rotmatrix(rotMatrix, curQuat);
            glMultMatrixf(&rotMatrix[0][0]);
	    glTranslatef(-centerOfWindow[0], -centerOfWindow[1], 
                                             -centerOfWindow[2]);
	    glPushMatrix();
	       glTranslatef(localTrans[0], localTrans[1], localTrans[2]);
	       glScalef(localScale[0], localScale[1], localScale[2]);
   }

   /** Function to be called after all GL API calls to finish the rendering */
   inline void processEnd()
   {
	       depthTest = glIsEnabled(GL_DEPTH_TEST);
	       glDisable(GL_DEPTH_TEST);

               UnitAxis::render(axisLength);

               switch(mTransType)
               {
	          case START_RUBBER_BAND:
		     createRubberBand(viewPort, mouseValues, projMatrix);
                  break;
	          case START_POLYLINE:
                     if( pointsList.size() > 3 )
		        createPolyline(pointsList);
                  break;
                  case RECT_SCALE:
                  case RECT_PICK:
                  case SCALE_DECR:
                  case SCALE_INCR:
                  case GLOBAL_SCALE:
                  {
		     // drawing the center of scaling
		     glColor3f(1.0, 0.0, 0.0);
		     glPointSize(15.0);
		     glEnable(GL_POINT_SMOOTH);
		     glBegin(GL_POINTS);
			//glVertex3dv(startOfScaling);
			glVertex3dv(centerOfScaling);
			//glVertex3dv(endOfScaling);
		     glEnd();
		     glPointSize(1.0);
		     glDisable(GL_POINT_SMOOTH);
                  }
                  break;
                  default:
                  break;
               }

	       glGetDoublev(GL_MODELVIEW_MATRIX, mvMatrix);
	       glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);

               if( depthTest )
                  glEnable(GL_DEPTH_TEST);

               glMatrixMode(GL_MODELVIEW);
            glPopMatrix(); //local scaling
         glPopMatrix(); //global rotations
      glPopMatrix(); //global translations
      mTransType = NO_TRANSFORM;
      if( onlyOnce & DISP_ON )
      {
         onlyOnce &= ~DISP_ON;
         computeWindowCenter();
      }
//cout << "END OF PROCESS\n";
//printMatrices();
   }

   /** Function for processing the mouse events in a standalone application
     * \param mouse_values startx, starty, endx and endy
     */
   void processMouseEvents(int mouse_values[4]);

   /** Setting new window width and height 
     * \param w Width of window
     * \param h height of window
     */
   void setViewport(int w, int h)
   {
      memset(viewPort, 0, sizeof(int) * 4);
      memcpy(viewPort+2, &w, sizeof(int));
      memcpy(viewPort+3, &h, sizeof(int));
cout << "VIEWPORT=" << viewPort[0] << "," << viewPort[1] << ","
     << viewPort[2] << "," << viewPort[3] << endl;
   }

   /** The model view mtrix and the projection matrix are stored here. This
     * function gets the pointer those arrays.
     * \param mv_mat Model view matrix
     * \param proj_mat projection matrix
     * \param view_port Viewport values x,y,width and height
     */
   void getMatrices(double **mv_mat, double **proj_mat, int **view_port)
   {
      *mv_mat = mvMatrix;
      *proj_mat = projMatrix;
      *view_port = viewPort;
   }

   /** Computes the center of window for scaling and rotation
     */
   void computeWindowCenter();
   /** Updates the center of window with the new model extents
     */
   void updateWindowCenter();

   /** For polyline pick a point is added at the clicked location
     * \param mouse_values startx, starty, endx and endy for the mouse location
     */
   void addPoint(float *mouse_values);

   /** Gets a pointer to the list of points added for polyline pick
     * \return Pointer the list of points
     */
   vector<Point3D *> *getPointsList()
   {
      return &pointsList;
   }

   /** Prints all the matrices stored in this class */
   void printMatrices();

   /** Saves the arrays stored in this class for restore at a later point time
     * \param file_ptr File pointer
     */
   void save(FILE *file_ptr);
   /** Restores the arrays stored in this class from save
     * \param file_ptr File pointer
     */
   void restore(FILE *file_ptr);

   /** Computes the global arrays for the given scale factors 
     * \param sx Scale in x direction
     * \param sy Scale in y direction
     * \param sz Scale in z direction
     */
   void updateScaleForWindow(float sx, float sy, float sz);

   /** moving the model to the center of window */
   void moveToCOW();

private:
   /** Updates the center of scaling for rectangular pick of entities */
   void updateRect();

   /** Updates the center of scaling for rubber band scaling of model */
   void updateLocalScaleForRubberBand();

   /** Updates the local scale for scaling to occur from center of window */
   void updateLocalScaleForWindow(float scale);

private:
   /** transformation type */
   int   mTransType, depthTest, onlyOnce;
   //float localRot[3];
   /** Coordinates of viewport x,y,w,h */
   int viewPort[4];
   /** pointer to model extents */
   float *modelExtents;
   /** Values of mouse location */
   float mouseValues[4];

   /** Local translation values */
   float localTrans[3];
   /** Global rotation values */
   float globalRot[3];
   /** Global translational values */
   float globalTrans[3];
   /** Local scaling values */
   float localScale[3];
   float axisLength[3];
   //float globalScale[3];
   /** Rotation matrix */
   float rotMatrix[4][4];
   float curQuat[4];
   /** Model view matrix of model */
   double mvMatrix[16];
   double mvMatrixLocal[16];
   /** Projection matrix of model */
   double projMatrix[16];
   /** Center for scaling of model */
   double centerOfScaling[3];
   /** Center of windows for scaling of model */
   double centerOfWindow[3];
   /** start point of scaling */
   double startOfScaling[3];
   /** end point of scaling */
   double endOfScaling[3];
   /** List of points for polygon pick */
   vector<Point3D *> pointsList;
};

#endif

