#ifndef _D3_PLOT_H_
#define _D3_PLOT_H_

#include "datatypes.h"
#include "bufio.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <iostream>
#include <map>
#include <vector>
#include "sphelement.h"
#include "statedata.h"
#include "getdb.h"

# define NS_SWAP16(x) ((((x) & 0xff) << 8) | (((x) >> 8) & 0xff))
# define NS_SWAP32(x) ((NS_SWAP16((x) & 0xffff) << 16) | (NS_SWAP16((x) >> 16)))

typedef enum _D3PlotTypes
{
   CONTROL_DATA,
   MATTYPE_DATA,
   FLUID_MATID_DATA,
   GEOMETRY_DATA,
   USERID_DATA,
   USER_NODEID_DATA,
   USER_SOLIDID_DATA,
   USER_BEAMID_DATA,
   USER_SHELLID_DATA,
   USER_TSHELLID_DATA,
   USER_ORDERED_MAT_DATA,
   USER_UNORDERED_MAT_DATA,
   USER_CROSS_REF_DATA,
   ADAPTED_ELEMENT_DATA,
   SPH_NODE_MAT_DATA,
   RIGID_ROAD_SURFACE_DATA,
   GLOBAL_STATE_DATA,
   NODE_STATE_DATA,
   CFD_STATE_DATA,
   ELEMENT_STATE_DATA,
   SOLID_STATE_DATA,
   TSHELL_STATE_DATA,
   BEAM_STATE_DATA,
   SHELL_STATE_DATA,
   ELEMENT_DELETE_STATE_DATA,
   SPH_STATE_DATA,
   ROAD_SURFACE_MOTION_STATE_DATA,

   D3PLOT_MAX_TYPE
}
  D3PlotTypes;

using namespace std;

typedef union
{
   unsigned int i;
   float f;
}
   fi_type;

typedef struct _RoadSurface
{
   WORD surfid;
   WORD surfnseg;
   WORD *surfsegs;
}
  RoadSurfaceStr, *RoadSurfacePtr;

typedef map<WORD, WORD> WWMap;
typedef pair<WORD, WORD> WWPair;
typedef map<WORD, WORD>::iterator WWIter;

class D3Plot
{
public:
   D3Plot(const char *results_dir, const char *user_dir, const char *model) :
     resultsDir(results_dir), userDir(user_dir), modelName(model), fileSize(0),
         fileDes(-1), fileBase(NULL), numTimes(0)
   {
       nsort = nsrh = nsrb = nsrs = nsrt = nsortd = nsorth = nsortb = nsorts = nsortt
               = nsrma = nsrmu = nsrmp = nsrtm = numrbs = nmmat = 0;
       nnode = nseg = nsurf = motion = 0;
       nel4 = 0;
       nel8 = 0;
       numnp = 0;
       nadapt = 0;
       numrw = 0;

       m_GetDB = new GetDB(model);
   }



   int openFile(const char *buf);
   int getNextAvailableFile(string & fileName, string & filePath, int *newDirFlag = NULL);

   int readControlData();

   int readMatTypeData();
   int readFluidMaterialIDData();
   int readSPHElementDataFlags();
   int readGeometryData(int dirIndex);
   int readUserIDsData();
   int readAdaptedElementParentList();
   int readSPHNodeMaterialList();
   int readRigidRoadSurfaceData();
   int readStateData(int dirIndex);
   int readCFDData(StateDataPtr state);
   int readElementData(StateDataPtr state);
   int readElementDeleteOption(StateDataPtr state);
   int readSPHElementData(StateDataPtr state);
   int readRoadSurfaceMotion(StateDataPtr state);

   int saveNodes(WORD file_idx, WORD stride, FWORD *node_values, WORD nnd, float timeStep);
   int saveGeometryNodes();
   int saveNodeIDs();

   int saveElements
       (
      const char *elm_name,
      WORD num_nodes,
      WORD num_elms,
      WORD *elm_ids,
           int elm_type
       );

   int writeStartFile(WORD file_idx, set<FWORD> & x, set<FWORD> & y, set<FWORD> & z);

   ~D3Plot()
   {
      int i, j;
      if( fileDes >= 0 )
      {
         munmap(fileBase, fileSize);
         close(fileDes);
      }

      if( nel8 )
         delete solids;
      if( nel4 )
         delete shells;
      if( nel2 )
      {
         delete beams;
          nel2 = 0;
      }
      if( nelt )
         delete tshells;
      if( numnp )
         delete nodeValues;

      //if( nsortd )
      //   delete nusern;
      if( nsorth )
         delete nuserh;
      if( nsortb )
         delete nuserb;
      if( nsorts )
         delete nusers;
      if( nsortt )
         delete nusert;
      if( nmsph )
     delete sphnodemat;
      if( nmmat )
      {
         delete norder;
         delete nsrmua;
         delete nsrmpa;
      }
      if( nadapt )
         delete adapt;
      if( nnode )
      {
         delete nodeids;
         delete surfnodes;

     if( nsurf )
     {
        for(i = 0; i < nsurf; i++)
           if( roads[i].surfnseg )
          delete roads[i].surfsegs;

        delete roads;
     }
      }

      if( numTimes )
      {
         for(i = 0; i < numTimes; i++)
         {
            if( numrw )
               delete states[i].global.rw_forces;
            if( total_mats )
            {
               delete states[i].global.mat.ie;
               delete states[i].global.mat.ke;
               delete states[i].global.mat.mass;
               delete states[i].global.mat.vel;
            }
            if( states[i].nnd )
               delete states[i].nodedata;
            if( ncfdv1 )
               delete states[i].cfdv1;
            if( ncfdv2 )
               delete states[i].cfdv2;
            if( states[i].solids )
               delete states[i].solids;
            if( states[i].tshells )
               delete states[i].tshells;
            if( states[i].shells )
               delete states[i].shells;
            if( states[i].beams )
               delete states[i].beams;
            if( states[i].num_del_opt )
               delete states[i].del_opt;
        if( nmsph )
            {
               for(j = 0; i < nmsph; j++)
               {
                  if( states[i].sph[j].stress )
                     delete states[i].sph[j].stress;
                  if( states[i].sph[j].strain )
                     delete states[i].sph[j].strain;
               }

           delete states[i].sph;
            }
         }
      }
   }

   int readData(int dirIndex);

   int getNextFile(string & fileName, string & filePath);

private:
   const  char *resultsDir, *userDir, *modelName;
   char   title[64], time[32], date[32], machine[32], code[32], releaseNum[32];
   size_t fileSize;
   int      fileDes;
   WORD    mdlopt, istrn, mattyp, mat_arb, useModelFlag, inum,srcVer, dynaVersion;
   WORD   *fileBase, *m_FilePtr;
   //CONTROL SECTION
   WORD   ndim, numnp, icode, nglbv, it, iu, iv, ia, nel8, nummat8, nv3d,
          nel2, nummat2, nv1d, nel4, nummat4, nv2d, neiph, neips, maxint,
          edlopt, nmsph, ngpsph, narbs, nelt, nummatt, nv3dt, ioshl1, ioshl2,
          ioshl3, ioshl4, ialemat, ncfdv1, ncfdv2, nadapt, numrw, total_mats,
          num_cfdv1, num_cfdv2, nDim, numds, numst;
   //SPH ELEMENT FLAGS
   WORD   isphfg[9];
   WORD   nummat, numrbe, *irbtyp, *fluidid;
   //GEOMETRIC DATA
   WORD   *solids, *shells, *tshells, *beams;
   //USER IDS DATA
   WORD   nsort, nsrh, nsrb, nsrs, nsrt, nsortd, nsorth, nsortb, nsorts,
          nsortt, nsrma, nsrmu, nsrmp, nsrtm, numrbs, nmmat;
   WORD   *nusern, *nuserh, *nuserb, *nusers, *nusert, *norder, *nsrmua,
          *nsrmpa, *adapt, *sphnodemat;
   FWORD  version;
   //NODE VALUES
   FWORD  *nodeValues;
   //RIGID ROAD SURFACE DATA
   WORD   nnode, nseg, nsurf, motion, *nodeids;
   FWORD  *surfnodes; //xyz values for each node
   RoadSurfacePtr roads;
   //STATE DATA
   WORD numTimes;
   StateDataPtr states;
   WWMap nidsMap;
   string  m_BaseFile;
   GetDB  *m_GetDB;
   static int m_GlobalFileIndex;
   float m_Exts[6];
};

/**
 * Return 1 if this is a little endian machine, 0 if big endian.
 */
static inline unsigned char mesa_little_endian(void)
{
   const unsigned int ui = 1; /* intentionally not static */
   return *((const unsigned char *) &ui);
}


#endif

